from __future__ import annotations

import base64
import hashlib
import weakref
from abc import ABC
from copy import deepcopy
from datetime import datetime, timezone, timedelta
from enum import Enum, unique
from functools import wraps, cached_property
from typing import Dict, Iterator, List, ClassVar, Optional, TypedDict, Any, TypeVar, Type, Callable, Set, Tuple
from collections import defaultdict

from attr import resolve_types
from attrs import define, field, Factory
from prometheus_client import Counter, Summary

from resotolib.json import from_json as _from_json, to_json as _to_json
from resotolib.logger import log
from resotolib.types import Json
from resotolib.utils import make_valid_timestamp, utc_str

metrics_resource_pre_cleanup_exceptions = Counter(
    "resource_pre_cleanup_exceptions_total",
    "Number of resource pre_cleanup() exceptions",
    ["cloud", "account", "region", "kind"],
)
metrics_resource_cleanup_exceptions = Counter(
    "resource_cleanup_exceptions_total",
    "Number of resource cleanup() exceptions",
    ["cloud", "account", "region", "kind"],
)
metrics_resource_cleanup = Summary("resoto_resource_cleanup_seconds", "Time it took the resource cleanup() method")


def unless_protected(f: Callable[..., bool]) -> Callable[..., bool]:
    @wraps(f)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> bool:
        if not isinstance(self, BaseResource):
            raise ValueError("unless_protected() only supports BaseResource type objects")
        if self.protected:
            log.error(f"Resource {self.rtdname} is protected - refusing modification")
            self.log(("Modification was requested even though resource is protected" " - refusing"))
            return False
        return f(self, *args, **kwargs)

    return wrapper


# To define predecessors and successors of a resource by kind
# Example:
# reference_kinds: ClassVar[ModelReference] = {
#   "successors": {"default": ["base"], "delete": ["other_kind"]]},
#   "predecessors": {"delete": ["other"]},
# }
class ModelReferenceEdges(TypedDict, total=False):
    default: List[str]
    delete: List[str]


class ModelReference(TypedDict, total=False):
    predecessors: ModelReferenceEdges
    successors: ModelReferenceEdges


class EdgeType(Enum):
    default = "default"
    delete = "delete"

    @staticmethod
    def from_value(value: Optional[str] = None) -> EdgeType:
        try:
            return EdgeType(value)
        except ValueError:
            pass
        return EdgeType.default


class ResourceChanges:
    def __init__(self, node: BaseResource) -> None:
        self.node = node
        self.reported: Set[str] = set()
        self.desired: Set[str] = set()
        self.metadata: Set[str] = set()
        self.changed = False

    def add(self, property: str) -> None:
        if property == "tags":
            self.reported.add(property)
        elif property == "clean":
            self.desired.add(property)
        elif property in ("cleaned", "protected"):
            self.metadata.add(property)
        elif property == "log":
            pass
        else:
            raise ValueError(f"Unknown property {property}")
        self.changed = True

    def get(self) -> Dict[str, Any]:
        changes: Dict[str, Any] = {}
        for section in ("reported", "desired", "metadata"):
            for attribute in getattr(self, section, []):
                if section not in changes:
                    changes[section] = {}
                try:
                    changes[section][attribute] = getattr(self.node, attribute)
                except AttributeError:
                    log.error(f"Resource {self.node.rtdname} has no attribute {attribute}")
        if len(self.node.event_log) > 0:
            if "metadata" not in changes:
                changes[section] = {}
            changes["metadata"]["event_log"] = self.node.str_event_log
        return changes


MetricName = str
StatName = str


@define(eq=False, slots=False, kw_only=True)
class BaseResource(ABC):
    """A BaseResource is any node we're connecting to the Graph()

    BaseResources have an id, name and tags. The id is a unique id used to search for
    the resource within the Graph. The name is used for display purposes. Tags are
    key/value pairs that get exported in the GRAPHML view.

    There's also class variables, kind, phantom and reference_kinds.
    `kind` is a string describing the type of resource, e.g. 'aws_ec2_instance'
       or 'some_cloud_load_balancer'.
    `phantom` is a bool describing whether the resource actually exists within
       the cloud or if it's just a phantom resource like pricing information
       or usage quota. I.e. some information relevant to the cloud account
       but not actually existing in the form of a usable resource.
    `reference_kinds` is a list of kinds that can be connected to this resource for
       the related edge type as successor or predecessor.
    """

    kind: ClassVar[str] = "resource"
    kind_display: ClassVar[str] = "Resource"
    kind_description: ClassVar[str] = "A generic resource."
    phantom: ClassVar[bool] = False
    reference_kinds: ClassVar[ModelReference] = {}
    metadata: ClassVar[Dict[str, Any]] = {"icon": "resource", "group": "misc"}

    id: str
    tags: Dict[str, Optional[str]] = Factory(dict)
    name: Optional[str] = field(default=None)
    _cloud: "Optional[BaseCloud]" = field(default=None, repr=False)
    _account: "Optional[BaseAccount]" = field(default=None, repr=False)
    _region: "Optional[BaseRegion]" = field(default=None, repr=False)
    _zone: "Optional[BaseZone]" = field(default=None, repr=False)
    _resotocore_id: Optional[str] = field(default=None, repr=False)
    _resotocore_revision: Optional[str] = field(default=None, repr=False)
    _resotocore_query_tag: Optional[str] = field(default=None, repr=False)
    _clean: bool = False
    _cleaned: bool = False
    _protected: bool = False
    _deferred_connections: List[Dict[str, Any]] = field(factory=list)
    _resource_usage: Dict[MetricName, Dict[StatName, float]] = field(factory=lambda: defaultdict(dict))
    _metadata: Dict[str, Any] = field(factory=dict)  # values will be exported in the metadata section of the node

    ctime: Optional[datetime] = field(
        default=None,
        metadata={"synthetic": {"age": "trafo.duration_to_datetime"}},
    )
    mtime: Optional[datetime] = field(
        default=None,
        metadata={"synthetic": {"last_update": "trafo.duration_to_datetime"}},
    )
    atime: Optional[datetime] = field(
        default=None,
        metadata={"synthetic": {"last_access": "trafo.duration_to_datetime"}},
    )

    def __attrs_post_init__(self) -> None:
        if self.name is None:
            self.name = self.id
        self._changes: ResourceChanges = ResourceChanges(self)
        self.__graph = None
        self.__log: List[Json] = []
        self._raise_tags_exceptions: bool = False
        if not hasattr(self, "_ctime"):
            self._ctime = None
        if not hasattr(self, "_atime"):
            self._atime = None
        if not hasattr(self, "_mtime"):
            self._mtime = None

    def _keys(self) -> Tuple[Any, ...]:
        """Return a tuple of all keys that make this resource unique

        Must not be called before the resource is connected to the graph
        as the relative location within the graph is used to determine the
        tuple of unique keys.

        E.g. instance -> aws -> 123457 -> us-east-1 -> us-east-1b -> i-987654 -> myServer
        """
        if self._graph is None:
            raise RuntimeError(f"_keys() called on {self.rtdname} before resource was added to graph")
        return self.kind, self.cloud().id, self.account().id, self.region().id, self.zone().id, self.id

    @property
    def safe_name(self) -> str:
        return self.name or self.id

    @property
    def dname(self) -> str:
        if self.id == self.name:
            return self.id
        return f"{self.name} ({self.id})"

    @property
    def kdname(self) -> str:
        return f"{self.kind} {self.dname}"

    rtdname = kdname

    @classmethod
    def from_json(cls: Type[BaseResourceType], json: Json) -> BaseResourceType:
        return _from_json(json, cls)

    def to_json(self) -> Json:
        return _to_json(self, strip_nulls=True)

    def log(self, msg: str, data: Optional[Any] = None, exception: Optional[Exception] = None) -> None:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        log_entry = {
            "timestamp": now,
            "msg": str(msg),
            "exception": repr(exception) if exception else None,
            "data": deepcopy(data),
        }
        self.__log.append(log_entry)
        self._changes.add("log")

    def add_change(self, change: str) -> None:
        self._changes.add(change)

    @property
    def changes(self) -> ResourceChanges:
        return self._changes

    @property
    def event_log(self) -> List[Json]:
        return self.__log

    @property
    def str_event_log(self) -> List[Json]:
        return [
            {
                "timestamp": utc_str(le["timestamp"]),
                "msg": le["msg"],
                "exception": le["exception"],
            }
            for le in self.__log
        ]

    def update_tag(self, key: str, value: str) -> bool:
        raise NotImplementedError

    def delete_tag(self, key: str) -> bool:
        raise NotImplementedError

    @cached_property
    def chksum(self) -> str:
        """Return a checksum of the resource."""
        return (
            base64.urlsafe_b64encode(hashlib.blake2b(str(self._keys()).encode(), digest_size=16).digest())
            .decode("utf-8")
            .rstrip("=")
        )

    @property
    def age(self) -> Optional[timedelta]:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        if self.ctime is not None:
            return now - self.ctime
        else:
            return None

    @property
    def last_access(self) -> Optional[timedelta]:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        if self.atime is not None:
            return now - self.atime
        else:
            return None

    @property
    def last_update(self) -> Optional[timedelta]:
        now = datetime.utcnow().replace(tzinfo=timezone.utc)
        if self.mtime is not None:
            return now - self.mtime
        else:
            return None

    def _ctime_getter(self) -> Optional[datetime]:
        if ctime_string := self.tags.get("resoto:ctime"):
            try:
                return make_valid_timestamp(datetime.fromisoformat(ctime_string))
            except ValueError:
                pass
        return self._ctime

    def _ctime_setter(self, value: Optional[datetime]) -> None:
        self._ctime = make_valid_timestamp(value) if value else None  # type: ignore

    def _atime_getter(self) -> Optional[datetime]:
        return self._atime

    def _atime_setter(self, value: Optional[datetime]) -> None:
        self._atime = make_valid_timestamp(value) if value else None  # type: ignore

    def _mtime_getter(self) -> Optional[datetime]:
        return self._mtime

    def _mtime_setter(self, value: Optional[datetime]) -> None:
        self._mtime = make_valid_timestamp(value) if value else None  # type: ignore

    @property
    def clean(self) -> bool:
        return self._clean

    @clean.setter
    @unless_protected
    def clean(self, value: bool) -> None:
        if self.phantom and value:
            raise ValueError(f"Can't cleanup phantom resource {self.rtdname}")

        clean_str = "" if value else "not "
        self.log(f"Setting to {clean_str}be cleaned")
        log.debug(f"Setting {self.rtdname} to {clean_str}be cleaned")
        self._changes.add("clean")
        self._clean = value

    @property
    def cleaned(self) -> bool:
        return self._cleaned

    @property
    def protected(self) -> bool:
        return self._protected

    @protected.setter
    def protected(self, value: bool) -> None:
        """Protects the resource from cleanup
        This property acts like a fuse, once protected it can't be unprotected
        """
        if self.protected:
            log.debug(f"Resource {self.rtdname} is already protected")
            return
        if value:
            log.debug(f"Protecting resource {self.rtdname}")
            self.log("Protecting resource")
            self._changes.add("protected")
            self._protected = value

    # deprecated. future collectors plugins should be responsible for running pre_cleanup
    # and calling delete_resource on resources
    @metrics_resource_cleanup.time()
    @unless_protected
    def cleanup(self, graph: Optional[Any] = None) -> bool:
        if self.phantom:
            raise RuntimeError(f"Can't cleanup phantom resource {self.rtdname}")

        if self.cleaned:
            log.info(f"Resource {self.rtdname} has already been cleaned up")
            return True

        self._changes.add("cleaned")
        if graph is None:
            graph = self._graph

        account = self.account(graph)
        region = self.region(graph)
        if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
            raise RuntimeError(f"Could not determine account or region for cleanup of {self.rtdname}")

        log_suffix = f" in account {account.dname} region {region.name}"
        self.log("Trying to clean up")
        log.info(f"Trying to clean up {self.rtdname}{log_suffix}")
        try:
            if deleted := self.delete(graph):
                self._cleaned = True
                self.log("Successfully cleaned up")
                log.info(f"Successfully cleaned up {self.rtdname}{log_suffix}")
        except Exception as e:
            self.log("An error occurred during clean up", exception=e)
            log.exception(f"An error occurred during clean up {self.rtdname}{log_suffix}")
            cloud = self.cloud(graph)
            metrics_resource_cleanup_exceptions.labels(
                cloud=cloud.name,
                account=account.dname,
                region=region.name,
                kind=self.kind,
            ).inc()
            raise
        if not deleted:
            raise RuntimeError(f"Failed to clean up {self.rtdname}{log_suffix}")
        return True

    # deprecated. future collectors plugins should be responsible for running pre_cleanup
    # and calling pre_delete_resource on resources
    @unless_protected
    def pre_cleanup(self, graph: Optional[Any] = None) -> bool:
        if not hasattr(self, "pre_delete"):
            return True

        if graph is None:
            graph = self._graph

        if self.phantom:
            raise RuntimeError(f"Can't cleanup phantom resource {self.rtdname}")

        if self.cleaned:
            log.debug(f"Resource {self.rtdname} has already been cleaned up")
            return True

        account = self.account(graph)
        region = self.region(graph)
        if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
            log.error(("Could not determine account or region for pre cleanup of" f" {self.rtdname}"))
            return False

        log_suffix = f" in account {account.dname} region {region.name}"
        self.log("Trying to run pre clean up")
        log.debug(f"Trying to run pre clean up {self.rtdname}{log_suffix}")
        try:
            if not getattr(self, "pre_delete")(graph):
                self.log("Failed to run pre clean up")
                log.error(f"Failed to run pre clean up {self.rtdname}{log_suffix}")
                return False
            self.log("Successfully ran pre clean up")
            log.info(f"Successfully ran pre clean up {self.rtdname}{log_suffix}")
        except Exception as e:
            self.log("An error occurred during pre clean up", exception=e)
            log.exception(f"An error occurred during pre clean up {self.rtdname}{log_suffix}")
            cloud = self.cloud(graph)
            metrics_resource_pre_cleanup_exceptions.labels(
                cloud=cloud.name,
                account=account.dname,
                region=region.name,
                kind=self.kind,
            ).inc()
            raise
        return True

    @unless_protected
    def delete(self, graph: Any) -> bool:
        return False

    def account(self, graph: Optional[Any] = None) -> "BaseAccount":
        account: Optional[BaseAccount] = None
        if graph is None:
            graph = self._graph
        if self._account:
            account = self._account
        elif graph:
            account = graph.search_first_parent_class(self, BaseAccount)
        if account is None:
            account = UnknownAccount(id="undefined", tags={})
        return account

    def cloud(self, graph: Optional[Any] = None) -> "BaseCloud":
        cloud: Optional[BaseCloud] = None
        if graph is None:
            graph = self._graph
        if self._cloud:
            cloud = self._cloud
        elif graph:
            cloud = graph.search_first_parent_class(self, BaseCloud)
        if cloud is None:
            cloud = UnknownCloud(id="undefined", tags={})
        return cloud

    def region(self, graph: Optional[Any] = None) -> "BaseRegion":
        region: Optional[BaseRegion] = None
        if graph is None:
            graph = self._graph
        if self._region:
            region = self._region
        elif graph:
            region = graph.search_first_parent_class(self, BaseRegion)
        if region is None:
            region = UnknownRegion(id="undefined", tags={})
        return region

    def zone(self, graph: Optional[Any] = None) -> "BaseZone":
        zone: Optional[BaseZone] = None
        if graph is None:
            graph = self._graph
        if self._zone:
            zone = self._zone
        elif graph:
            zone = graph.search_first_parent_class(self, BaseZone)
        if zone is None:
            zone = UnknownZone(id="undefined", tags={})
        return zone

    def resource_location(self, graph: Optional[Any] = None) -> "BaseResource":
        if graph is None:
            graph = self._graph
        zone = self.zone(graph)
        if zone.name != "undefined":
            return zone
        region = self.region(graph)
        if region.name != "undefined":
            return region
        account = self.account(graph)
        if account.name != "undefined":
            return account
        cloud = self.cloud(graph)
        if cloud.name != "undefined":
            return cloud
        return UnknownLocation(id="undefined", tags={})

    def add_deferred_connection(
        self, search: Dict[str, Any], parent: bool = True, edge_type: EdgeType = EdgeType.default
    ) -> None:
        self._deferred_connections.append({"search": search, "parent": parent, "edge_type": edge_type})

    def resolve_deferred_connections(self, graph: Any) -> None:
        if graph is None:
            graph = self._graph
        while self._deferred_connections:
            dc = self._deferred_connections.pop(0)
            node = graph.search_first_all(dc["search"])
            edge_type = dc["edge_type"]
            if node:
                if dc["parent"]:
                    src = node
                    dst = self
                else:
                    src = self
                    dst = node
                graph.add_edge(src, dst, edge_type=edge_type)

    def predecessors(self, graph: Optional[Any] = None, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        """Returns an iterator of the node's parent nodes"""
        if graph is None:
            graph = self._graph
        if graph is None:
            return iter(())
        return graph.predecessors(self, edge_type=edge_type)  # type: ignore

    def successors(self, graph: Optional[Any] = None, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        """Returns an iterator of the node's child nodes"""
        if graph is None:
            graph = self._graph
        if graph is None:
            return iter(())
        return graph.successors(self, edge_type=edge_type)  # type: ignore

    def predecessor_added(self, resource: BaseResource, graph: Any) -> None:
        """Called when a predecessor is added to this node"""
        pass

    def successor_added(self, resource: BaseResource, graph: Any) -> None:
        """Called when a successor is added to this node"""
        pass

    def ancestors(self, graph: Any, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        """Returns an iterator of the node's ancestors"""
        if graph is None:
            graph = self._graph
        if graph is None:
            return iter(())
        return graph.ancestors(self, edge_type=edge_type)  # type: ignore

    def descendants(self, graph: Any, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        """Returns an iterator of the node's descendants"""
        if graph is None:
            graph = self._graph
        if graph is None:
            return iter(())
        return graph.descendants(self, edge_type=edge_type)  # type: ignore

    @property
    def _graph(self) -> Optional[Any]:
        if self.__graph is not None:
            return self.__graph()
        else:
            return None

    @_graph.setter
    def _graph(self, value: Any) -> None:
        self.__graph = weakref.ref(value)  # type: ignore

    def __getstate__(self) -> Dict[str, Any]:
        ret = self.__dict__.copy()
        ret["_BaseResource__graph"] = None
        return ret

    def __setstate__(self, state: Dict[str, Any]) -> None:
        self.__dict__.update(state)


BaseResource.ctime = property(BaseResource._ctime_getter, BaseResource._ctime_setter)  # type: ignore
BaseResource.mtime = property(BaseResource._mtime_getter, BaseResource._mtime_setter)  # type: ignore
BaseResource.atime = property(BaseResource._atime_getter, BaseResource._atime_setter)  # type: ignore


BaseResourceType = TypeVar("BaseResourceType", bound=BaseResource)


@define(eq=False, slots=False)
class PhantomBaseResource(BaseResource):
    kind: ClassVar[str] = "phantom_resource"
    kind_display: ClassVar[str] = "Phantom Resource"
    kind_description: ClassVar[str] = "A generic phantom resource."
    phantom: ClassVar[bool] = True

    def update_tag(self, key: str, value: str) -> bool:
        log.error(f"Resource {self.rtdname} is a phantom resource and does not maintain tags")
        return False

    def delete_tag(self, key: str) -> bool:
        log.error(f"Resource {self.rtdname} is a phantom resource and does not maintain tags")
        return False

    def delete(self, graph: Any) -> bool:
        log.error(f"Resource {self.rtdname} is a phantom resource and can't be deleted")
        return False

    def cleanup(self, graph: Optional[Any] = None) -> bool:
        log.error(f"Resource {self.rtdname} is a phantom resource and can't be cleaned up")
        return False


@define(eq=False, slots=False)
class BaseQuota(PhantomBaseResource):
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "control"}

    kind: ClassVar[str] = "quota"
    kind_display: ClassVar[str] = "Quota"
    kind_description: ClassVar[str] = "A service quota."
    quota: Optional[float] = None
    usage: Optional[float] = None
    quota_type: Optional[str] = None

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if self.quota is not None:
            self.quota = float(self.quota)
        if self.usage is not None:
            self.usage = float(self.usage)

    @property
    def usage_percentage(self) -> float:
        if self.quota is not None and self.usage is not None and self.quota > 0.0:
            return self.usage / self.quota * 100
        else:
            return 0.0


@define(eq=False, slots=False)
class BaseType(BaseQuota):
    kind: ClassVar[str] = "type"
    kind_display: ClassVar[str] = "Type"
    kind_description: ClassVar[str] = "A generic type."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "type", "group": "control"}


@define(eq=False, slots=False)
class BaseInstanceQuota(BaseQuota):
    kind: ClassVar[str] = "instance_quota"
    kind_display: ClassVar[str] = "Instance Quota"
    kind_description: ClassVar[str] = "An instance quota."
    instance_type: Optional[str] = None
    metadata: ClassVar[Dict[str, Any]] = {"icon": "instance", "group": "compute"}

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.instance_type = self.id
        self.quota_type = "standard"


@define(eq=False, slots=False)
class BaseInstanceType(BaseType):
    kind: ClassVar[str] = "instance_type"
    kind_display: ClassVar[str] = "Instance Type"
    kind_description: ClassVar[str] = "An instance type."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "instance_type", "group": "compute"}
    instance_type: Optional[str] = None
    instance_cores: float = 0.0
    instance_memory: float = 0.0
    ondemand_cost: Optional[float] = None
    reservations: Optional[int] = None

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        if self.instance_type is None:
            self.instance_type = self.id
        if self.reservations is not None:
            self.reservations = int(self.reservations)
        if self.ondemand_cost is not None:
            self.ondemand_cost = float(self.ondemand_cost)


@define(eq=False, slots=False)
class BaseCloud(BaseResource):
    kind: ClassVar[str] = "base_cloud"
    kind_display: ClassVar[str] = "Cloud"
    kind_description: ClassVar[str] = "A cloud."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "cloud", "group": "control"}

    def cloud(self, graph: Optional[Any] = None) -> BaseCloud:
        return self


@define(eq=False, slots=False)
class BaseAccount(BaseResource):
    kind: ClassVar[str] = "account"
    kind_display: ClassVar[str] = "Account"
    kind_description: ClassVar[str] = "An account."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "account", "group": "control"}

    def account(self, graph: Optional[Any] = None) -> BaseAccount:
        return self


@define(eq=False, slots=False)
class BaseRegion(BaseResource):
    kind: ClassVar[str] = "region"
    kind_display: ClassVar[str] = "Region"
    kind_description: ClassVar[str] = "A region."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "region", "group": "control"}

    def _keys(self) -> Tuple[Any, ...]:
        if self._graph is None:
            raise RuntimeError(f"_keys() called on {self.rtdname} before resource was added to graph")
        return self.kind, self.cloud().id, self.account().id, self.region().id, self.zone().id, self.id, self.name

    def region(self, graph: Optional[Any] = None) -> BaseRegion:
        return self


@define(eq=False, slots=False)
class BaseZone(BaseResource):
    kind: ClassVar[str] = "zone"
    kind_display: ClassVar[str] = "Zone"
    kind_description: ClassVar[str] = "A zone."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "zone", "group": "control"}

    def zone(self, graph: Optional[Any] = None) -> BaseZone:
        return self


class InstanceStatus(Enum):
    RUNNING = "running"
    STOPPED = "stopped"
    TERMINATED = "terminated"
    BUSY = "busy"
    UNKNOWN = "unknown"


@define(eq=False, slots=False)
class BaseInstance(BaseResource):
    kind: ClassVar[str] = "instance"
    kind_display: ClassVar[str] = "Instance"
    kind_description: ClassVar[str] = "An instance."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "instance", "group": "compute"}
    instance_cores: float = 0.0
    instance_memory: float = 0.0
    instance_type: Optional[str] = ""
    instance_status: Optional[InstanceStatus] = None


@define(eq=False, slots=False)
class BaseVolumeType(BaseType):
    kind: ClassVar[str] = "volume_type"
    kind_display: ClassVar[str] = "Volume Type"
    kind_description: ClassVar[str] = "A volume type."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "volume_type", "group": "storage"}
    volume_type: str = ""
    ondemand_cost: Optional[float] = None

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.volume_type = self.id
        if self.ondemand_cost is not None:
            self.ondemand_cost = float(self.ondemand_cost)


@unique
class VolumeStatus(Enum):
    IN_USE = "in-use"
    AVAILABLE = "available"
    BUSY = "busy"
    ERROR = "error"
    DELETED = "deleted"
    UNKNOWN = "unknown"


@define(eq=False, slots=False)
class BaseNetworkShare(BaseResource, ABC):
    kind: ClassVar[str] = "network_share"
    kind_display: ClassVar[str] = "Network Share"
    kind_description: ClassVar[str] = "A network share."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "network_share", "group": "storage"}
    share_size: int = 0
    share_type: str = ""
    share_status: Optional[str] = None
    share_iops: Optional[int] = None
    share_throughput: Optional[float] = None  # bytes per second
    share_encrypted: Optional[bool] = None


@define(eq=False, slots=False)
class BaseVolume(BaseResource):
    kind: ClassVar[str] = "volume"
    kind_display: ClassVar[str] = "Volume"
    kind_description: ClassVar[str] = "A volume."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "volume", "group": "storage"}
    volume_size: int = 0
    volume_type: str = ""
    volume_status: Optional[VolumeStatus] = None
    volume_iops: Optional[int] = None
    volume_throughput: Optional[int] = None
    volume_encrypted: Optional[bool] = None
    snapshot_before_delete: bool = False


@define(eq=False, slots=False)
class BaseSnapshot(BaseResource):
    kind: ClassVar[str] = "snapshot"
    kind_display: ClassVar[str] = "Snapshot"
    kind_description: ClassVar[str] = "A snapshot."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "snapshot", "group": "storage"}
    snapshot_status: str = ""
    description: Optional[str] = None
    volume_id: Optional[str] = None
    volume_size: int = 0
    encrypted: bool = False
    owner_id: Optional[str] = None
    owner_alias: Optional[str] = None


@define(eq=False, slots=False)
class Cloud(BaseCloud):
    kind: ClassVar[str] = "cloud"
    kind_display: ClassVar[str] = "Cloud"
    kind_description: ClassVar[str] = "A cloud."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "cloud", "group": "control"}

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class GraphRoot(PhantomBaseResource):
    kind: ClassVar[str] = "graph_root"
    kind_display: ClassVar[str] = "Graph Root"
    kind_description: ClassVar[str] = "The root of the graph."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "graph_root", "group": "control"}

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class BaseBucket(BaseResource):
    kind: ClassVar[str] = "bucket"
    kind_display: ClassVar[str] = "Storage Bucket"
    kind_description: ClassVar[str] = "A storage bucket."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "bucket", "group": "storage"}


@define(eq=False, slots=False)
class BaseServerlessFunction(BaseResource):
    kind: ClassVar[str] = "serverless_function"
    kind_display: ClassVar[str] = "Serverless Function"
    kind_description: ClassVar[str] = "A serverless function."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "function", "group": "compute"}


@define(eq=False, slots=False)
class BaseKeyPair(BaseResource):
    kind: ClassVar[str] = "keypair"
    kind_display: ClassVar[str] = "Key Pair"
    kind_description: ClassVar[str] = "A key pair."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "keypair", "group": "access_control"}
    fingerprint: str = ""


@define(eq=False, slots=False)
class BaseBucketQuota(BaseQuota):
    kind: ClassVar[str] = "bucket_quota"
    kind_display: ClassVar[str] = "Bucket Quota"
    kind_description: ClassVar[str] = "A bucket quota."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "storage"}


@define(eq=False, slots=False)
class BaseNetwork(BaseResource):
    kind: ClassVar[str] = "network"
    kind_display: ClassVar[str] = "Network"
    kind_description: ClassVar[str] = "A network."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "network", "group": "networking"}


@define(eq=False, slots=False)
class BaseNetworkQuota(BaseQuota):
    kind: ClassVar[str] = "network_quota"
    kind_display: ClassVar[str] = "Network Quota"
    kind_description: ClassVar[str] = "A network quota."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "networking"}


@define(eq=False, slots=False)
class BaseFirewall(BaseResource):
    kind: ClassVar[str] = "firewall"
    kind_display: ClassVar[str] = "Firewall"
    kind_description: ClassVar[str] = "A firewall."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "firewall", "group": "networking"}


@define(eq=False, slots=False)
class BaseDatabase(BaseResource):
    kind: ClassVar[str] = "database"
    kind_display: ClassVar[str] = "Database"
    kind_description: ClassVar[str] = "A database."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "database", "group": "database"}
    db_type: str = ""
    db_status: str = ""
    db_endpoint: Optional[str] = ""
    db_version: Optional[str] = None
    db_publicly_accessible: Optional[bool] = None
    instance_type: str = ""
    volume_size: int = 0
    volume_iops: Optional[int] = None
    volume_encrypted: Optional[bool] = None


@define(eq=False, slots=False)
class BaseLoadBalancer(BaseResource):
    kind: ClassVar[str] = "load_balancer"
    kind_display: ClassVar[str] = "Load Balancer"
    kind_description: ClassVar[str] = "A load balancer."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "load_balancer", "group": "networking"}
    lb_type: str = ""
    public_ip_address: Optional[str] = None
    backends: List[str] = field(factory=list)


@define(eq=False, slots=False)
class BaseLoadBalancerQuota(BaseQuota):
    kind: ClassVar[str] = "load_balancer_quota"
    kind_display: ClassVar[str] = "Load Balancer Quota"
    kind_description: ClassVar[str] = "A load balancer quota."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "networking"}


@define(eq=False, slots=False)
class BaseSubnet(BaseResource):
    kind: ClassVar[str] = "subnet"
    kind_display: ClassVar[str] = "Subnet"
    kind_description: ClassVar[str] = "A subnet."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "subnet", "group": "networking"}


@define(eq=False, slots=False)
class BaseGateway(BaseResource):
    kind: ClassVar[str] = "gateway"
    kind_display: ClassVar[str] = "Gateway"
    kind_description: ClassVar[str] = "A gateway."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "gateway", "group": "networking"}


@define(eq=False, slots=False)
class BaseTunnel(BaseResource):
    kind: ClassVar[str] = "tunnel"
    kind_display: ClassVar[str] = "Networking Tunnel"
    kind_description: ClassVar[str] = "A networking tunnel."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "tunnel", "group": "networking"}


@define(eq=False, slots=False)
class BaseGatewayQuota(BaseQuota):
    kind: ClassVar[str] = "gateway_quota"
    kind_display: ClassVar[str] = "Gateway Quota"
    kind_description: ClassVar[str] = "A gateway quota."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "networking"}


@define(eq=False, slots=False)
class BaseSecurityGroup(BaseResource):
    kind: ClassVar[str] = "security_group"
    kind_display: ClassVar[str] = "Security Group"
    kind_description: ClassVar[str] = "A security group."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "security_group", "group": "networking"}


@define(eq=False, slots=False)
class BaseRoutingTable(BaseResource):
    kind: ClassVar[str] = "routing_table"
    kind_display: ClassVar[str] = "Routing Table"
    kind_description: ClassVar[str] = "A routing table."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "routing_table", "group": "networking"}


@define(eq=False, slots=False)
class BaseNetworkAcl(BaseResource):
    kind: ClassVar[str] = "network_acl"
    kind_display: ClassVar[str] = "Network ACL"
    kind_description: ClassVar[str] = "A network access control list."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "acl", "group": "networking"}


@define(eq=False, slots=False)
class BasePeeringConnection(BaseResource):
    kind: ClassVar[str] = "peering_connection"
    kind_display: ClassVar[str] = "Peering Connection"
    kind_description: ClassVar[str] = "A peering connection."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "connection", "group": "networking"}


@define(eq=False, slots=False)
class BaseEndpoint(BaseResource):
    kind: ClassVar[str] = "endpoint"
    kind_display: ClassVar[str] = "Endpoint"
    kind_description: ClassVar[str] = "An endpoint."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "endpoint", "group": "networking"}


@define(eq=False, slots=False)
class BaseNetworkInterface(BaseResource):
    kind: ClassVar[str] = "network_interface"
    kind_display: ClassVar[str] = "Network Interface"
    kind_description: ClassVar[str] = "A network interface."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "network_interface", "group": "networking"}
    network_interface_status: str = ""
    network_interface_type: str = ""
    mac: str = ""
    private_ips: List[str] = field(factory=list)
    public_ips: List[str] = field(factory=list)
    v6_ips: List[str] = field(factory=list)
    description: str = ""


@define(eq=False, slots=False)
class BaseUser(BaseResource):
    kind: ClassVar[str] = "user"
    kind_display: ClassVar[str] = "User"
    kind_description: ClassVar[str] = "A user."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "user", "group": "access_control"}


@define(eq=False, slots=False)
class BaseGroup(BaseResource):
    kind: ClassVar[str] = "group"
    kind_display: ClassVar[str] = "Group"
    kind_description: ClassVar[str] = "A group."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "group", "group": "access_control"}


@define(eq=False, slots=False)
class BasePolicy(BaseResource):
    kind: ClassVar[str] = "policy"
    kind_display: ClassVar[str] = "Policy"
    kind_description: ClassVar[str] = "A policy."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "policy", "group": "access_control"}


@define(eq=False, slots=False)
class BaseRole(BaseResource):
    kind: ClassVar[str] = "role"
    kind_display: ClassVar[str] = "Role"
    kind_description: ClassVar[str] = "A role."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "role", "group": "access_control"}


@define(eq=False, slots=False)
class BaseInstanceProfile(BaseResource):
    kind: ClassVar[str] = "instance_profile"
    kind_display: ClassVar[str] = "Instance Profile"
    kind_description: ClassVar[str] = "An instance profile."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "instance_profile", "group": "access_control"}


@define(eq=False, slots=False)
class BaseAccessKey(BaseResource):
    kind: ClassVar[str] = "access_key"
    kind_display: ClassVar[str] = "Access Key"
    kind_description: ClassVar[str] = "An access key."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "key", "group": "access_control"}
    access_key_status: str = ""


@define(eq=False, slots=False)
class BaseCertificate(BaseResource):
    kind: ClassVar[str] = "certificate"
    kind_display: ClassVar[str] = "Certificate"
    kind_description: ClassVar[str] = "A certificate."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "certificate", "group": "access_control"}
    expires: Optional[datetime] = None
    dns_names: Optional[List[str]] = None
    sha1_fingerprint: Optional[str] = None


@define(eq=False, slots=False)
class BaseCertificateQuota(BaseQuota):
    kind: ClassVar[str] = "certificate_quota"
    kind_display: ClassVar[str] = "Certificate Quota"
    kind_description: ClassVar[str] = "A certificate quota."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "quota", "group": "access_control"}


@define(eq=False, slots=False)
class BaseStack(BaseResource):
    kind: ClassVar[str] = "stack"
    kind_display: ClassVar[str] = "Stack"
    kind_description: ClassVar[str] = "A stack."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "stack", "group": "control"}
    stack_status: str = ""
    stack_status_reason: str = ""
    stack_parameters: Dict[str, str] = field(factory=dict)


@define(eq=False, slots=False)
class BaseAutoScalingGroup(BaseResource):
    kind: ClassVar[str] = "autoscaling_group"
    kind_display: ClassVar[str] = "Auto Scaling Group"
    kind_description: ClassVar[str] = "An auto scaling group."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "autoscaling_group", "group": "compute"}
    min_size: int = -1
    max_size: int = -1


@define(eq=False, slots=False)
class BaseIPAddress(BaseResource):
    kind: ClassVar[str] = "ip_address"
    kind_display: ClassVar[str] = "IP Address"
    kind_description: ClassVar[str] = "An IP address."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "network_address", "group": "networking"}
    ip_address: str = ""
    ip_address_family: str = ""


@define(eq=False, slots=False)
class BaseHealthCheck(BaseResource):
    kind: ClassVar[str] = "health_check"
    kind_display: ClassVar[str] = "Health Check"
    kind_description: ClassVar[str] = "A health check."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "health_check", "group": "compute"}
    check_interval: int = -1
    healthy_threshold: int = -1
    unhealthy_threshold: int = -1
    timeout: int = -1
    health_check_type: str = ""


@define(eq=False, slots=False)
class BaseDNSZone(BaseResource):
    kind: ClassVar[str] = "dns_zone"
    kind_display: ClassVar[str] = "DNS Zone"
    kind_description: ClassVar[str] = "A DNS zone."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "dns", "group": "networking"}


@define(eq=False, slots=False)
class BaseDNSRecordSet(BaseResource):
    kind: ClassVar[str] = "dns_record_set"
    kind_display: ClassVar[str] = "DNS Record Set"
    kind_description: ClassVar[str] = "A DNS record set."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "dns", "group": "networking"}

    record_ttl: Optional[int] = None
    record_type: str = ""
    record_values: List[str] = field(factory=list)

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.record_type = self.record_type.upper()

    def dns_zone(self, graph: Optional[Any] = None) -> "BaseDNSZone":
        if graph is None:
            graph = self._graph
        dns_zone = graph.search_first_parent_class(self, BaseDNSZone) if graph else None
        return dns_zone or UnknownDNSZone(id="undefined", tags={})

    def _keys(self) -> tuple[Any, ...]:
        if self._graph is None:
            raise RuntimeError(f"_keys() called on {self.rtdname} before resource was added to graph")
        return (
            self.kind,
            self.cloud().id,
            self.account().id,
            self.region().id,
            self.zone().id,
            self.dns_zone().id,
            self.id,
            self.name,
            self.record_type,
        )


@define(eq=False, slots=False)
class BaseDNSRecord(BaseResource):
    kind: ClassVar[str] = "dns_record"
    kind_display: ClassVar[str] = "DNS Record"
    kind_description: ClassVar[str] = "A DNS record."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "dns_record", "group": "networking"}

    record_ttl: int = -1
    record_type: str = ""
    record_data: str = ""
    record_value: str = ""
    record_priority: Optional[int] = None
    record_port: Optional[int] = None
    record_weight: Optional[int] = None
    record_flags: Optional[int] = None
    record_tag: Optional[str] = None
    record_mname: Optional[str] = None
    record_rname: Optional[str] = None
    record_serial: Optional[int] = None
    record_refresh: Optional[int] = None
    record_retry: Optional[int] = None
    record_expire: Optional[int] = None
    record_minimum: Optional[int] = None

    def __attrs_post_init__(self) -> None:
        super().__attrs_post_init__()
        self.record_type = self.record_type.upper()

    def dns_zone(self, graph: Optional[Any] = None) -> "BaseDNSZone":
        if graph is None:
            graph = self._graph
        dns_zone = graph.search_first_parent_class(self, BaseDNSZone) if graph else None
        return dns_zone or UnknownDNSZone(id="undefined", tags={})

    def _keys(self) -> tuple[Any, ...]:
        if self._graph is None:
            raise RuntimeError(f"_keys() called on {self.rtdname} before resource was added to graph")
        return (
            self.kind,
            self.cloud().id,
            self.account().id,
            self.region().id,
            self.zone().id,
            self.dns_zone().id,
            self.id,
            self.name,
            self.record_type,
            self.record_data,
        )


@define(eq=False, slots=False)
class BaseOrganizationalRoot(BaseResource):
    kind: ClassVar[str] = "organizational_root"
    kind_display: ClassVar[str] = "Organizational Root"
    kind_description: ClassVar[str] = "An Organizational Root."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "resource", "group": "misc"}


@define(eq=False, slots=False)
class BaseOrganizationalUnit(BaseResource):
    kind: ClassVar[str] = "organizational_unit"
    kind_display: ClassVar[str] = "Organizational Unit"
    kind_description: ClassVar[str] = "An Organizational Unit."
    metadata: ClassVar[Dict[str, Any]] = {"icon": "resource", "group": "misc"}


@define(eq=False, slots=False)
class UnknownCloud(BaseCloud):
    kind: ClassVar[str] = "unknown_cloud"
    kind_display: ClassVar[str] = "Unknown Cloud"
    kind_description: ClassVar[str] = "An unknown cloud."

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class UnknownAccount(BaseAccount):
    kind: ClassVar[str] = "unknown_account"
    kind_display: ClassVar[str] = "Unknown Account"
    kind_description: ClassVar[str] = "An unknown account."

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class UnknownRegion(BaseRegion):
    kind: ClassVar[str] = "unknown_region"
    kind_display: ClassVar[str] = "Unknown Region"
    kind_description: ClassVar[str] = "An unknown region."

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class UnknownDNSZone(BaseDNSZone):
    kind: ClassVar[str] = "unknown_dns_zone"
    kind_display: ClassVar[str] = "Unknown DNS Zone"
    kind_description: ClassVar[str] = "An unknown DNS zone."

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class UnknownZone(BaseZone):
    kind: ClassVar[str] = "unknown_zone"
    kind_display: ClassVar[str] = "Unknown Zone"
    kind_description: ClassVar[str] = "An unknown zone."

    def delete(self, graph: Any) -> bool:
        return False


@define(eq=False, slots=False)
class UnknownLocation(BaseResource):
    kind: ClassVar[str] = "unknown_location"
    kind_display: ClassVar[str] = "Unknown Location"
    kind_description: ClassVar[str] = "An unknown location."

    def delete(self, graph: Any) -> bool:
        return False


resolve_types(BaseResource)  # noqa



import os
import tempfile
from resotolib.x509 import (
    gen_rsa_key,
    gen_csr,
    bootstrap_ca,
    sign_csr,
    write_csr_to_file,
    write_cert_to_file,
    write_key_to_file,
    load_csr_from_file,
    load_cert_from_file,
    load_key_from_file,
    key_to_bytes,
    cert_fingerprint,
    cert_is_signed_by_ca,
)


def test_x509():
    with tempfile.TemporaryDirectory() as tmp:
        ca_key, ca_cert = bootstrap_ca()
        cert_key = gen_rsa_key()
        gen_csr(cert_key)  # dummy call to generate CSR without SANs
        cert_csr = gen_csr(
            cert_key,
            san_dns_names=["example.com"],
            san_ip_addresses=["10.0.1.1", "10.0.0.0/24"],
        )
        cert_crt = sign_csr(cert_csr, ca_key, ca_cert)
        ca_key_path = os.path.join(tmp, "ca.key")
        ca_cert_path = os.path.join(tmp, "ca.crt")

        cert_key_path = os.path.join(tmp, "cert.key")
        cert_key_passphrase = "foobar"
        cert_csr_path = os.path.join(tmp, "cert.csr")
        cert_crt_path = os.path.join(tmp, "cert.crt")

        write_key_to_file(ca_key, key_path=ca_key_path)
        write_cert_to_file(ca_cert, cert_path=ca_cert_path)

        write_key_to_file(cert_key, key_path=cert_key_path, passphrase=cert_key_passphrase)
        write_csr_to_file(cert_csr, csr_path=cert_csr_path)
        write_cert_to_file(cert_crt, cert_path=cert_crt_path)

        loaded_ca_key = load_key_from_file(ca_key_path)
        loaded_ca_cert = load_cert_from_file(ca_cert_path)
        loaded_cert_key = load_key_from_file(cert_key_path, passphrase=cert_key_passphrase)
        loaded_cert_csr = load_csr_from_file(cert_csr_path)
        loaded_cert_crt = load_cert_from_file(cert_crt_path)

        assert loaded_ca_cert == ca_cert
        assert loaded_cert_csr == cert_csr
        assert loaded_cert_crt == cert_crt
        assert cert_fingerprint(loaded_ca_cert) == cert_fingerprint(ca_cert)
        assert cert_fingerprint(loaded_cert_crt) == cert_fingerprint(cert_crt)
        assert key_to_bytes(ca_key) == key_to_bytes(loaded_ca_key)
        assert key_to_bytes(cert_key) == key_to_bytes(loaded_cert_key)
        assert cert_is_signed_by_ca(cert_crt, ca_cert)



import sys
import os
import unittest

from resotolib.tree import Tree, Node, NodeIDAbsentError

"""
NOTE:
This module is copied and ported to proper Python 3.0 from https://github.com/caesar0301/treelib
Since this library has outdated vulnerable dependencies, we decided to copy it here directly.
"""


class TreeCase(unittest.TestCase):
    def setUp(self):
        tree = Tree(identifier="tree 1")
        tree.create_node("Harry", "harry")
        tree.create_node("Jane", "jane", parent="harry")
        tree.create_node("Bill", "bill", parent="harry")
        tree.create_node("Diane", "diane", parent="jane")
        tree.create_node("George", "george", parent="bill")
        # Harry
        #   |-- Jane
        #       |-- Diane
        #   |-- Bill
        #       |-- George
        self.tree = tree
        self.copytree = Tree(self.tree, deep=True)

    @staticmethod
    def get_t1():
        """
        root
        ├── A
        │   └── A1
        └── B
        """
        t = Tree(identifier="t1")
        t.create_node(tag="root", identifier="r")
        t.create_node(tag="A", identifier="a", parent="r")
        t.create_node(tag="B", identifier="b", parent="r")
        t.create_node(tag="A1", identifier="a1", parent="a")
        return t

    @staticmethod
    def get_t2():
        """
        root2
        ├── C
        └── D
            └── D1
        """
        t = Tree(identifier="t2")
        t.create_node(tag="root2", identifier="r2")
        t.create_node(tag="C", identifier="c", parent="r2")
        t.create_node(tag="D", identifier="d", parent="r2")
        t.create_node(tag="D1", identifier="d1", parent="d")
        return t

    def test_tree(self):
        self.assertEqual(isinstance(self.tree, Tree), True)
        self.assertEqual(isinstance(self.copytree, Tree), True)

    def test_is_root(self):
        # retro-compatibility
        self.assertTrue(self.tree._nodes["harry"].is_root())
        self.assertFalse(self.tree._nodes["jane"].is_root())

    def test_tree_wise_is_root(self):
        subtree = self.tree.subtree("jane", identifier="subtree 2")
        # harry is root of tree 1 but not present in subtree 2
        self.assertTrue(self.tree._nodes["harry"].is_root("tree 1"))
        self.assertNotIn("harry", subtree._nodes)
        # jane is not root of tree 1 but is root of subtree 2
        self.assertFalse(self.tree._nodes["jane"].is_root("tree 1"))
        self.assertTrue(subtree._nodes["jane"].is_root("subtree 2"))

    def test_paths_to_leaves(self):
        paths = self.tree.paths_to_leaves()
        self.assertEqual(len(paths), 2)
        self.assertTrue(["harry", "jane", "diane"] in paths)
        self.assertTrue(["harry", "bill", "george"] in paths)

    def test_nodes(self):
        self.assertEqual(len(self.tree.nodes), 5)
        self.assertEqual(len(self.tree.all_nodes()), 5)
        self.assertEqual(self.tree.size(), 5)
        self.assertEqual(self.tree.get_node("jane").tag, "Jane")
        self.assertEqual(self.tree.contains("jane"), True)
        self.assertEqual("jane" in self.tree, True)
        self.assertEqual(self.tree.contains("alien"), False)
        self.tree.create_node("Alien", "alien", parent="jane")
        self.assertEqual(self.tree.contains("alien"), True)
        self.tree.remove_node("alien")

    def test_getitem(self):
        """Nodes can be accessed via getitem."""
        for node_id in self.tree.nodes:
            try:
                self.tree[node_id]
            except NodeIDAbsentError:
                self.fail("Node access should be possible via getitem.")
        try:
            self.tree["root"]
        except NodeIDAbsentError:
            pass
        else:
            self.fail("There should be no default fallback value for getitem.")

    def test_parent(self):
        for nid in self.tree.nodes:
            if nid == self.tree.root:
                self.assertEqual(self.tree.parent(nid), None)
            else:
                self.assertEqual(self.tree.parent(nid) in self.tree.all_nodes(), True)

    def test_ancestor(self):
        for nid in self.tree.nodes:
            if nid == self.tree.root:
                self.assertEqual(self.tree.ancestor(nid), None)
            else:
                for level in range(self.tree.level(nid) - 1, 0, -1):
                    self.assertEqual(self.tree.ancestor(nid, level=level) in self.tree.all_nodes(), True)

    def test_children(self):
        for nid in self.tree.nodes:
            children = self.tree.is_branch(nid)
            for child in children:
                self.assertEqual(self.tree[child] in self.tree.all_nodes(), True)
            children = self.tree.children(nid)
            for child in children:
                self.assertEqual(child in self.tree.all_nodes(), True)
        try:
            self.tree.is_branch("alien")
        except NodeIDAbsentError:
            pass
        else:
            self.fail("The absent node should be declaimed.")

    def test_remove_node(self):
        self.tree.create_node("Jill", "jill", parent="george")
        self.tree.create_node("Mark", "mark", parent="jill")
        self.assertEqual(self.tree.remove_node("jill"), 2)
        self.assertEqual(self.tree.get_node("jill") is None, True)
        self.assertEqual(self.tree.get_node("mark") is None, True)

    def test_tree_wise_depth(self):
        # Try getting the level of this tree
        self.assertEqual(self.tree.depth(), 2)
        self.tree.create_node("Jill", "jill", parent="george")
        self.assertEqual(self.tree.depth(), 3)
        self.tree.create_node("Mark", "mark", parent="jill")
        self.assertEqual(self.tree.depth(), 4)

        # Try getting the level of the node
        """
        Harry
        |___ Bill
        |    |___ George
        |         |___ Jill
        |              |___ Mark
        |___ Jane
        |    |___ Diane
        """
        self.assertEqual(self.tree.depth(self.tree.get_node("mark")), 4)
        self.assertEqual(self.tree.depth(self.tree.get_node("jill")), 3)
        self.assertEqual(self.tree.depth(self.tree.get_node("george")), 2)
        self.assertEqual(self.tree.depth("jane"), 1)
        self.assertEqual(self.tree.depth("bill"), 1)
        self.assertEqual(self.tree.depth("harry"), 0)

        # Try getting Exception
        node = Node("Test One", "identifier 1")
        self.assertRaises(NodeIDAbsentError, self.tree.depth, node)

        # Reset the test case
        self.tree.remove_node("jill")

    def test_leaves(self):
        # retro-compatibility
        leaves = self.tree.leaves()
        for nid in self.tree.expand_tree():
            self.assertEqual((self.tree[nid].is_leaf()) == (self.tree[nid] in leaves), True)
        leaves = self.tree.leaves(nid="jane")
        for nid in self.tree.expand_tree(nid="jane"):
            self.assertEqual(self.tree[nid].is_leaf() == (self.tree[nid] in leaves), True)

    def test_tree_wise_leaves(self):
        leaves = self.tree.leaves()
        for nid in self.tree.expand_tree():
            self.assertEqual((self.tree[nid].is_leaf("tree 1")) == (self.tree[nid] in leaves), True)
        leaves = self.tree.leaves(nid="jane")
        for nid in self.tree.expand_tree(nid="jane"):
            self.assertEqual(self.tree[nid].is_leaf("tree 1") == (self.tree[nid] in leaves), True)

    def test_link_past_node(self):
        self.tree.create_node("Jill", "jill", parent="harry")
        self.tree.create_node("Mark", "mark", parent="jill")
        self.assertEqual("mark" not in self.tree.is_branch("harry"), True)
        self.tree.link_past_node("jill")
        self.assertEqual("mark" in self.tree.is_branch("harry"), True)

    def test_expand_tree(self):
        # default config
        # Harry
        #   |-- Jane
        #       |-- Diane
        #   |-- Bill
        #       |-- George
        # Traverse in depth first mode preserving insertion order
        nodes = [nid for nid in self.tree.expand_tree(sorting=False)]
        self.assertEqual(nodes, ["harry", "jane", "diane", "bill", "george"])
        self.assertEqual(len(nodes), 5)

        # By default, traverse depth first and sort child nodes by node tag
        nodes = [nid for nid in self.tree.expand_tree()]
        self.assertEqual(nodes, ["harry", "bill", "george", "jane", "diane"])
        self.assertEqual(len(nodes), 5)

        # expanding from specific node
        nodes = [nid for nid in self.tree.expand_tree(nid="bill")]
        self.assertEqual(nodes, ["bill", "george"])
        self.assertEqual(len(nodes), 2)

        # changing into width mode preserving insertion order
        nodes = [nid for nid in self.tree.expand_tree(mode=Tree.WIDTH, sorting=False)]
        self.assertEqual(nodes, ["harry", "jane", "bill", "diane", "george"])
        self.assertEqual(len(nodes), 5)

        # Breadth first mode, child nodes sorting by tag
        nodes = [nid for nid in self.tree.expand_tree(mode=Tree.WIDTH)]
        self.assertEqual(nodes, ["harry", "bill", "jane", "george", "diane"])
        self.assertEqual(len(nodes), 5)

        # expanding by filters
        # Stops at root
        nodes = [nid for nid in self.tree.expand_tree(filter_fn=lambda x: x.tag == "Bill")]
        self.assertEqual(len(nodes), 0)
        nodes = [nid for nid in self.tree.expand_tree(filter_fn=lambda x: x.tag != "Bill")]
        self.assertEqual(nodes, ["harry", "jane", "diane"])
        self.assertEqual(len(nodes), 3)

    def test_move_node(self):
        diane_parent = self.tree.parent("diane")
        self.tree.move_node("diane", "bill")
        self.assertEqual("diane" in self.tree.is_branch("bill"), True)
        self.tree.move_node("diane", diane_parent.identifier)

    def test_paste_tree(self):
        new_tree = Tree()
        new_tree.create_node("Jill", "jill")
        new_tree.create_node("Mark", "mark", parent="jill")
        self.tree.paste("jane", new_tree)
        self.assertEqual("jill" in self.tree.is_branch("jane"), True)
        self.tree.remove_node("jill")
        self.assertNotIn("jill", self.tree.nodes.keys())
        self.assertNotIn("mark", self.tree.nodes.keys())

    def test_merge(self):
        # merge on empty initial tree
        t1 = Tree(identifier="t1")
        t2 = self.get_t2()
        t1.merge(nid=None, new_tree=t2)

        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r2")
        self.assertEqual(set(t1._nodes.keys()), {"r2", "c", "d", "d1"})

        # merge empty new_tree (on root)
        t1 = self.get_t1()
        t2 = Tree(identifier="t2")
        t1.merge(nid="r", new_tree=t2)

        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertEqual(set(t1._nodes.keys()), {"r", "a", "a1", "b"})

        # merge at root
        t1 = self.get_t1()
        t2 = self.get_t2()
        t1.merge(nid="r", new_tree=t2)

        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertNotIn("r2", t1._nodes.keys())
        self.assertEqual(set(t1._nodes.keys()), {"r", "a", "a1", "b", "c", "d", "d1"})

        # merge on node
        t1 = self.get_t1()
        t2 = self.get_t2()
        t1.merge(nid="b", new_tree=t2)
        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertNotIn("r2", t1._nodes.keys())
        self.assertEqual(set(t1._nodes.keys()), {"r", "a", "a1", "b", "c", "d", "d1"})

    def test_paste(self):
        # paste under root
        t1 = self.get_t1()
        t2 = self.get_t2()
        t1.paste(nid="r", new_tree=t2)
        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertEqual(t1.parent("r2").identifier, "r")
        self.assertEqual(set(t1._nodes.keys()), {"r", "r2", "a", "a1", "b", "c", "d", "d1"})

        # paste under non-existing node
        t1 = self.get_t1()
        t2 = self.get_t2()
        with self.assertRaises(NodeIDAbsentError) as e:
            t1.paste(nid="not_existing", new_tree=t2)
        self.assertEqual(e.exception.args[0], "Node 'not_existing' is not in the tree")

        # paste under None nid
        t1 = self.get_t1()
        t2 = self.get_t2()
        with self.assertRaises(ValueError) as e:
            t1.paste(nid=None, new_tree=t2)
        self.assertEqual(e.exception.args[0], 'Must define "nid" under which new tree is pasted.')

        # paste under node
        t1 = self.get_t1()
        t2 = self.get_t2()
        t1.paste(nid="b", new_tree=t2)
        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertEqual(t1.parent("b").identifier, "r")
        self.assertEqual(set(t1._nodes.keys()), {"r", "a", "a1", "b", "c", "d", "d1", "r2"})
        # paste empty new_tree (under root)
        t1 = self.get_t1()
        t2 = Tree(identifier="t2")
        t1.paste(nid="r", new_tree=t2)

        self.assertEqual(t1.identifier, "t1")
        self.assertEqual(t1.root, "r")
        self.assertEqual(set(t1._nodes.keys()), {"r", "a", "a1", "b"})

    def test_rsearch(self):
        for nid in ["harry", "jane", "diane"]:
            self.assertEqual(nid in self.tree.rsearch("diane"), True)

    def test_subtree(self):
        subtree_copy = Tree(self.tree.subtree("jane"), deep=True)
        self.assertEqual(subtree_copy.parent("jane") is None, True)
        subtree_copy["jane"].tag = "Sweeti"
        self.assertEqual(self.tree["jane"].tag == "Jane", True)
        self.assertEqual(subtree_copy.level("diane"), 1)
        self.assertEqual(subtree_copy.level("jane"), 0)
        self.assertEqual(self.tree.level("jane"), 1)

    def test_remove_subtree(self):
        subtree_shallow = self.tree.remove_subtree("jane")
        self.assertEqual("jane" not in self.tree.is_branch("harry"), True)
        self.tree.paste("harry", subtree_shallow)

    def test_remove_subtree_whole_tree(self):
        self.tree.remove_subtree("harry")
        self.assertIsNone(self.tree.root)
        self.assertEqual(len(self.tree.nodes.keys()), 0)

    def test_to_json(self):
        self.tree.to_json()
        self.tree.to_json(True)

    def test_siblings(self):
        self.assertEqual(len(self.tree.siblings("harry")) == 0, True)
        self.assertEqual(self.tree.siblings("jane")[0].identifier == "bill", True)

    def test_tree_data(self):
        class Flower(object):
            def __init__(self, color):
                self.color = color

        self.tree.create_node("Jill", "jill", parent="jane", data=Flower("white"))
        self.assertEqual(self.tree["jill"].data.color, "white")
        self.tree.remove_node("jill")

    def test_show_data_property(self):
        new_tree = Tree()

        sys.stdout = open(os.devnull, "w")  # stops from printing to console

        try:

            class Flower(object):
                def __init__(self, color):
                    self.color = color

            new_tree.create_node("Jill", "jill", data=Flower("white"))
        finally:
            sys.stdout.close()
            sys.stdout = sys.__stdout__  # stops from printing to console

    def test_level(self):
        self.assertEqual(self.tree.level("harry"), 0)
        depth = self.tree.depth()
        self.assertEqual(self.tree.level("diane"), depth)
        self.assertEqual(self.tree.level("diane", lambda x: x.identifier != "jane"), depth - 1)

    def test_size(self):
        self.assertEqual(self.tree.size(level=2), 2)
        self.assertEqual(self.tree.size(level=1), 2)
        self.assertEqual(self.tree.size(level=0), 1)

    def tearDown(self):
        self.tree = None
        self.copytree = None

    def test_all_nodes_itr(self):
        """
        tests: Tree.all_nodes_iter
        Added by: William Rusnack
        """
        new_tree = Tree()
        self.assertEqual(len(new_tree.all_nodes_itr()), 0)
        nodes = list()
        nodes.append(new_tree.create_node("root_node"))
        nodes.append(new_tree.create_node("second", parent=new_tree.root))
        for nd in new_tree.all_nodes_itr():
            self.assertTrue(nd in nodes)

    def test_filter_nodes(self):
        """
        tests: Tree.filter_nodes
        Added by: William Rusnack
        """
        new_tree = Tree(identifier="tree 1")

        self.assertEqual(tuple(new_tree.filter_nodes(lambda n: True)), ())

        nodes = list()
        nodes.append(new_tree.create_node("root_node"))
        nodes.append(new_tree.create_node("second", parent=new_tree.root))

        self.assertEqual(tuple(new_tree.filter_nodes(lambda n: False)), ())
        self.assertEqual(tuple(new_tree.filter_nodes(lambda n: n.is_root("tree 1"))), (nodes[0],))
        self.assertEqual(tuple(new_tree.filter_nodes(lambda n: not n.is_root("tree 1"))), (nodes[1],))
        self.assertTrue(set(new_tree.filter_nodes(lambda n: True)), set(nodes))

    def test_loop(self):
        tree = Tree()
        tree.create_node("a", "a")
        tree.create_node("b", "b", parent="a")
        tree.create_node("c", "c", parent="b")
        tree.create_node("d", "d", parent="c")
        try:
            tree.move_node("b", "d")
        except ValueError:
            pass

    def test_modify_node_identifier_directly_failed(self):
        tree = Tree()
        tree.create_node("Harry", "harry")
        tree.create_node("Jane", "jane", parent="harry")
        n = tree.get_node("jane")
        self.assertTrue(n.identifier == "jane")

        # Failed to modify
        n.identifier = "xyz"
        self.assertTrue(tree.get_node("xyz") is None)
        self.assertTrue(tree.get_node("jane").identifier == "xyz")

    def test_modify_node_identifier_recursively(self):
        tree = Tree()
        tree.create_node("Harry", "harry")
        tree.create_node("Jane", "jane", parent="harry")
        n = tree.get_node("jane")
        self.assertTrue(n.identifier == "jane")

        # Success to modify
        tree.update_node(n.identifier, identifier="xyz")
        self.assertTrue(tree.get_node("jane") is None)
        self.assertTrue(tree.get_node("xyz").identifier == "xyz")

    def test_modify_node_identifier_root(self):
        tree = Tree(identifier="tree 3")
        tree.create_node("Harry", "harry")
        tree.create_node("Jane", "jane", parent="harry")
        tree.update_node(tree["harry"].identifier, identifier="xyz", tag="XYZ")
        self.assertTrue(tree.root == "xyz")
        self.assertTrue(tree["xyz"].tag == "XYZ")
        self.assertEqual(tree.parent("jane").identifier, "xyz")

    def test_subclassing(self):
        class SubNode(Node):
            pass

        class SubTree(Tree):
            node_class = SubNode

        tree = SubTree()
        node = tree.create_node()
        self.assertTrue(isinstance(node, SubNode))

        tree = Tree(node_class=SubNode)
        node = tree.create_node()
        self.assertTrue(isinstance(node, SubNode))

    def test_shallow_copy_hermetic_pointers(self):
        # tree 1
        # Harry
        #   └── Jane
        #       └── Diane
        #   └── Bill
        #       └── George
        tree2 = self.tree.subtree(nid="jane", identifier="tree 2")
        # tree 2
        # Jane
        #   └── Diane

        # check that in shallow copy, instances are the same
        self.assertIs(self.tree["jane"], tree2["jane"])
        self.assertEqual(self.tree["jane"]._predecessor, {"tree 1": "harry", "tree 2": None})
        self.assertEqual(dict(self.tree["jane"]._successors), {"tree 1": ["diane"], "tree 2": ["diane"]})

        # when creating new node on subtree, check that it has no impact on initial tree
        tree2.create_node("Jill", "jill", parent="diane")
        self.assertIn("jill", tree2)
        self.assertIn("jill", tree2.is_branch("diane"))
        self.assertNotIn("jill", self.tree)
        self.assertNotIn("jill", self.tree.is_branch("diane"))

    def test_paste_duplicate_nodes(self):
        t1 = Tree()
        t1.create_node(identifier="A")
        t2 = Tree()
        t2.create_node(identifier="A")
        t2.create_node(identifier="B", parent="A")

        with self.assertRaises(ValueError) as e:
            t1.paste("A", t2)
        self.assertEqual(e.exception.args, ("Duplicated nodes ['A'] exists.",))

    def test_shallow_paste(self):
        t1 = Tree()
        n1 = t1.create_node(identifier="A")

        t2 = Tree()
        n2 = t2.create_node(identifier="B")

        t3 = Tree()
        n3 = t3.create_node(identifier="C")

        t1.paste(n1.identifier, t2)
        self.assertEqual(t1.to_dict(), {"A": {"children": ["B"]}})
        t1.paste(n1.identifier, t3)
        self.assertEqual(t1.to_dict(), {"A": {"children": ["B", "C"]}})

        self.assertEqual(t1.level(n1.identifier), 0)
        self.assertEqual(t1.level(n2.identifier), 1)
        self.assertEqual(t1.level(n3.identifier), 1)

    def test_root_removal(self):
        t = Tree()
        t.create_node(identifier="root-A")
        self.assertEqual(len(t.nodes.keys()), 1)
        self.assertEqual(t.root, "root-A")
        t.remove_node(identifier="root-A")
        self.assertEqual(len(t.nodes.keys()), 0)
        self.assertEqual(t.root, None)
        t.create_node(identifier="root-B")
        self.assertEqual(len(t.nodes.keys()), 1)
        self.assertEqual(t.root, "root-B")



from __future__ import annotations

import json
import re
import tempfile
import threading
from collections import defaultdict, namedtuple, deque
from datetime import datetime
from enum import Enum
from functools import lru_cache
from time import time
from typing import Dict, Iterator, List, Tuple, Optional, Union, Any, Type, TypeVar, Set

import networkx
from attr import resolve_types
from attrs import define, fields
from networkx.algorithms.dag import is_directed_acyclic_graph
from prometheus_client import Summary
from typeguard import check_type

from resotolib.baseresources import (
    BaseCloud,
    BaseAccount,
    GraphRoot,
    Cloud,
    BaseResource,
    EdgeType,
    BaseRegion,
    BaseZone,
    MetricName,
    StatName,
)
from resotolib.core.model_export import (
    node_to_dict,
    dataclasses_to_resotocore_model,
    dynamic_object_to_resotocore_model,
)
from resotolib.json import to_json_str
from resotolib.logger import log
from resotolib.types import Json
from resotolib.utils import get_resource_attributes, unset_cached_properties, utc_str

T = TypeVar("T")


@define
class BySearchCriteria:
    query: str


@define
class ByNodeId:
    value: str


NodeSelector = Union[ByNodeId, BySearchCriteria]


metrics_graph_search = Summary("resoto_graph_search_seconds", "Time it took the Graph search() method")
metrics_graph_searchall = Summary("resoto_graph_searchall_seconds", "Time it took the Graph searchall() method")
metrics_graph_searchre = Summary("resoto_graph_searchre_seconds", "Time it took the Graph searchre() method")
metrics_graph_search_first = Summary(
    "resoto_graph_search_first_seconds",
    "Time it took the Graph search_first() method",
)
metrics_graph_search_first_all = Summary(
    "resoto_graph_search_first_all_seconds",
    "Time it took the Graph search_first_all() method",
)
metrics_graph_search_first_parent_class = Summary(
    "resoto_graph_search_first_parent_seconds",
    "Time it took the Graph search_first_parent_class() method",
)
metrics_graph_resolve_deferred_connections = Summary(
    "resoto_graph_resolve_deferred_connections",
    "Time it took the Graph resolve_deferred_connections() method",
)
metrics_graphcache_update_cache = Summary(
    "resoto_graphcache_update_cache_seconds",
    "Time it took the GraphCache update_cache() method",
)
metrics_graph2json = Summary("resoto_graph2json_seconds", "Time it took the graph2json() method")
metrics_graph2text = Summary("resoto_graph2text_seconds", "Time it took the graph2text() method")
metrics_graph2graphml = Summary("resoto_graph2graphml_seconds", "Time it took the graph2graphml() method")
metrics_graph2pickle = Summary("resoto_graph2pickle_seconds", "Time it took the graph2pickle() method")
metrics_graph2gexf = Summary("resoto_graph2gexf_seconds", "Time it took the graph2gexf() method")
metrics_graph2pajek = Summary("resoto_graph2pajek_seconds", "Time it took the graph2pajek() method")

EdgeKey = namedtuple("EdgeKey", ["src", "dst", "edge_type"])


class Graph(networkx.MultiDiGraph):  # type: ignore
    """A directed Graph"""

    def __init__(self, *args: Any, root: Optional[BaseResource] = None, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.root = None
        self._log_edge_creation = True
        if isinstance(root, BaseResource):
            self.root = root
            self.add_node(self.root, label=self.root.name, **get_resource_attributes(self.root))
        self.deferred_edges: List[Tuple[NodeSelector, NodeSelector, EdgeType]] = []

    def merge(self, graph: Graph, skip_deferred_edges: bool = False) -> None:
        """Merge another graph into ourselves

        If the other graph has a graph.root an edge will be created between
        it and our own graph root.
        """
        if isinstance(self.root, BaseResource) and isinstance(getattr(graph, "root", None), BaseResource):
            log.debug(f"Merging graph of {graph.root.kdname} into graph of {self.root.kdname}")  # type: ignore
            self.add_edge(self.root, graph.root)  # type: ignore
        else:
            log.warning("Merging graphs with no valid roots")

        try:
            self._log_edge_creation = False
            self.update(edges=graph.edges, nodes=graph.nodes)
            self.deferred_edges.extend(graph.deferred_edges)
        finally:
            self._log_edge_creation = True
        if not skip_deferred_edges:
            self.resolve_deferred_connections()

    def add_resource(
        self,
        parent: BaseResource,
        node_for_adding: BaseResource,
        edge_type: Optional[EdgeType] = None,
        **attr: Any,
    ) -> None:
        """Add a resource node to the graph

        When adding resource nodes to the graph there's always a label and a
        kind as well as an edge connecting the new resource with its parent
        resource. This way we should never have disconnected nodes within the graph.

        The graph_attributes are a Dict of key=value pairs that contain all the
        attributes of a graph node. When working with the graph that information isn't
        too important as each node is a standard Python object containing all its
        properties. However when exporting in graphml, gexf or json those attributes are
        the only thing being exported with the node name. So we'll try to turn every
        public property of a graph node into key=value attribute pairs to be exported
        with the graph.

        Attributes can also be used in graph searches to find nodes matching certain
        attributes.
        """
        resource_attr = get_resource_attributes(node_for_adding)

        self.add_node(node_for_adding, label=node_for_adding.name, **resource_attr, **attr)
        self.add_edge(src=parent, dst=node_for_adding, edge_type=edge_type)

    def add_node(self, node_for_adding: BaseResource, **attr: Any) -> None:
        super().add_node(node_for_adding, **attr)
        if isinstance(node_for_adding, BaseResource):
            # We hand a reference to ourselves to the added BaseResource
            # which stores it as a weakref.
            node_for_adding._graph = self

    def has_edge(
        self, src: BaseResource, dst: BaseResource, key: Optional[EdgeKey] = None, edge_type: Optional[EdgeType] = None
    ) -> bool:
        edge_type = edge_type or EdgeType.default
        key = key or EdgeKey(src=src, dst=dst, edge_type=edge_type)
        return super().has_edge(src, dst, key=key)  # type: ignore

    def add_edge(
        self,
        src: BaseResource,
        dst: BaseResource,
        key: Optional[EdgeKey] = None,
        edge_type: Optional[EdgeType] = None,
        **attr: Any,
    ) -> Optional[EdgeKey]:
        if src is None or dst is None:
            log.error(f"Not creating edge from or to NoneType: {src} to {dst}")
            return None

        if edge_type is None:
            edge_type = EdgeType.default
        if key is None:
            key = EdgeKey(src=src, dst=dst, edge_type=edge_type)

        if self.has_edge(src, dst, key=key):
            log.debug(f"Edge from {src} to {dst} already exists in graph")
            return None
        return_key = super().add_edge(src, dst, key=key, **attr)
        if self._log_edge_creation and isinstance(src, BaseResource) and isinstance(dst, BaseResource):
            log.debug(f"Added edge from {src.kdname} to {dst.kdname} (type: {edge_type.value})")
            try:
                src.successor_added(dst, self)
            except Exception:
                log.exception(
                    (f"Unhandled exception while telling {src.kdname}" f" that {dst.kdname} was added as a successor")
                )
            try:
                dst.predecessor_added(src, self)
            except Exception:
                log.exception(
                    (f"Unhandled exception while telling {dst.kdname}" f" that {src.kdname} was added as a predecessor")
                )
        return return_key  # type: ignore

    def add_deferred_edge(self, src: NodeSelector, dst: NodeSelector, edge_type: EdgeType = EdgeType.default) -> None:
        self.deferred_edges.append((src, dst, edge_type))

    def remove_node(self, node: BaseResource) -> None:
        super().remove_node(node)

    def remove_edge(
        self,
        src: BaseResource,
        dst: BaseResource,
        key: Optional[EdgeKey] = None,
        edge_type: Optional[EdgeType] = None,
    ) -> None:
        if edge_type is None:
            edge_type = EdgeType.default
        if key is None:
            key = EdgeKey(src=src, dst=dst, edge_type=edge_type)
        super().remove_edge(src, dst, key=key)

    def predecessors(self, node: BaseResource, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        if edge_type is None:
            edge_type = EdgeType.default
        for predecessor in super().predecessors(node):
            key = EdgeKey(predecessor, node, edge_type)
            if self.has_edge(predecessor, node, key=key):
                yield predecessor

    def successors(self, node: BaseResource, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        if edge_type is None:
            edge_type = EdgeType.default
        for successor in super().successors(node):
            key = EdgeKey(node, successor, edge_type)
            if self.has_edge(node, successor, key=key):
                yield successor

    def ancestors(self, node: BaseResource, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        return networkx.algorithms.dag.ancestors(self.edge_type_subgraph(edge_type), node)  # type: ignore

    def descendants(self, node: BaseResource, edge_type: Optional[EdgeType] = None) -> Iterator[BaseResource]:
        return networkx.algorithms.dag.descendants(self.edge_type_subgraph(edge_type), node)  # type: ignore

    def edge_type_subgraph(self, edge_type: Optional[EdgeType] = None) -> networkx.Graph:
        if edge_type is None:
            edge_type = EdgeType.default
        edges = []
        for edge in self.edges(keys=True):
            if len(edge) == 3:
                key: EdgeKey = edge[2]
                if key.edge_type == edge_type:
                    edges.append(edge)
        return self.edge_subgraph(edges)

    def is_acyclic_per_edge_type(self) -> bool:
        """
        Checks if the graph is acyclic with respect to each edge type.
        This means it is valid if there are cycles in the graph but not for the same edge type.
        :return: True if the graph is acyclic for all edge types, otherwise False.
        """
        log.debug("Ensuring graph is directed and acyclic per edge type")
        return True if self.find_cycle() is None else False

    def find_cycle(self) -> Optional[List[EdgeKey]]:
        """
        Returns the first cycle found in the graph per edge type.
        :return: first cycle found in the graph per edge type, or None if no cycles are found.
        """
        edges_per_type = defaultdict(list)
        for edge in self.edges(keys=True):
            if len(edge) == 3:
                key: EdgeKey = edge[2]
                edges_per_type[key.edge_type].append(edge)
        for edges in edges_per_type.values():
            typed_graph = self.edge_subgraph(edges)
            if not is_directed_acyclic_graph(typed_graph):
                return [edge[2] for edge in networkx.algorithms.cycles.find_cycle(typed_graph)]
        return None

    @metrics_graph_search.time()
    def search(self, attr: str, value: Any, regex_search: bool = False) -> Iterator[BaseResource]:
        """Search for graph nodes by their attribute value"""
        if value is None:
            log.debug(f"Not searching graph for nodes with attribute values {attr}: {value}")
            return ()
        log.debug((f"Searching graph for nodes with attribute values {attr}: {value}" f" (regex: {regex_search})"))
        for node in self.nodes():
            node_attr = getattr(node, attr, None)
            if (
                node_attr is not None
                and not callable(node_attr)
                and (
                    (regex_search is False and node_attr == value)
                    or (regex_search is True and re.search(value, str(node_attr)))
                )
            ):
                yield node

    @metrics_graph_searchre.time()
    def searchre(self, attr: str, regex: str) -> Iterator[BaseResource]:
        """Regex search for graph nodes by their attribute value"""
        log.debug(f"Regex searching graph for nodes with attribute values {attr}: {regex}")
        yield from self.search(attr, regex, regex_search=True)

    @metrics_graph_searchall.time()
    def searchall(self, match: Dict[str, Any]) -> Iterator[BaseResource]:
        """Search for graph nodes by multiple attributes and values"""
        for node in self.nodes():
            if all(getattr(node, attr, None) == value for attr, value in match.items()):
                yield node

    @metrics_graph_search_first.time()
    def search_first(self, attr: str, value: Any) -> Optional[BaseResource]:
        """Return the first graph node that matches a certain attribute value"""
        node = next(iter(self.search(attr, value)), None)
        if node:
            log.debug(f"Found node {node} with {attr}: {value}")
        else:
            log.debug(f"Found no node with {attr}: {value}")
        return node  # type

    @metrics_graph_search_first_all.time()
    def search_first_all(self, match: Dict[str, Any]) -> Optional[BaseResource]:
        """Return the first graph node that matches multiple attributes and values"""
        node = next(iter(self.searchall(match)), None)
        if node:
            log.debug(f"Found node {node} with {match}")
        else:
            log.debug(f"Found no node with {match}")
        return node

    @metrics_graph_search_first_parent_class.time()
    def search_first_parent_class(self, node: BaseResource, cls: Type[T]) -> Optional[T]:
        """Return the first parent node matching a certain class

        This is being used to search up the graph and e.g. find the account that the
        graph node is a member of.
        """
        queue = deque(self.predecessors(node))
        already_checked = set(self.predecessors(node))
        while queue:
            current = queue.popleft()
            if isinstance(current, cls):
                return current
            for n in self.predecessors(current):
                if n not in already_checked:
                    already_checked.add(n)
                    queue.append(n)
        return None

    @metrics_graph_resolve_deferred_connections.time()
    def resolve_deferred_connections(self) -> None:
        log.debug("Resolving deferred graph connections")
        for node in self.nodes:
            if isinstance(node, BaseResource):
                node.resolve_deferred_connections(self)

    def export_model(self, **kwargs: Any) -> List[Json]:
        return export_model(graph=self, **kwargs)

    def export_iterator(self) -> GraphExportIterator:
        return GraphExportIterator(self)


def resource_classes_to_resotocore_model(classes: Set[Type[Any]], **kwargs: Any) -> List[Json]:
    model = {c["fqn"]: c for c in dataclasses_to_resotocore_model(classes, **kwargs)}
    # create a phantom resource which defines default property paths
    dynamic_types: Dict[str, Dict[str, Type[Any]]] = {
        "resource_short_property_access": {
            "cloud": Union[Cloud, str],  # type: ignore
            "account": Union[BaseAccount, str],  # type: ignore
            "region": Union[BaseRegion, str],  # type: ignore
            "zone": Union[BaseZone, str],  # type: ignore
            "usage": Dict[MetricName, Dict[StatName, float]],
        }
    }
    for name, dynamic in dynamic_types.items():
        for dyn in dynamic_object_to_resotocore_model(name, dynamic, aggregate_root=True, traverse_dependant=False):
            model[dyn["fqn"]] = dyn
    return list(model.values())


def export_model(graph: Optional[Graph] = None, **kwargs: Any) -> List[Json]:
    """Return the graph node dataclass model in resotocore format"""
    classes = set()
    if graph is None:
        classes.add(BaseResource)
    else:
        for node in graph.nodes:
            classes.add(type(node))

    model = resource_classes_to_resotocore_model(classes, aggregate_root=BaseResource, with_description=False, **kwargs)
    for resource_model in model:
        if resource_model.get("fqn") == "resource":
            resource_model.get("properties", []).append(
                {
                    "name": "kind",
                    "kind": "string",
                    "required": True,
                    "description": "",
                }
            )
    return model


@lru_cache(maxsize=4096)  # Only resolve types once per type
def resolve_type(clazz: Type[Any]) -> None:
    resolve_types(clazz)


def validate_dataclass(node: BaseResource) -> bool:
    resolve_type(type(node))  # make sure all type annotations are resolved
    for field in fields(type(node)):
        value = getattr(node, field.name)
        try:
            check_type(value, field.type)
        except Exception:
            log.error(
                f"In {node.kdname} expected {field.name}"
                f" type {field.type} ({type(field.type)})"
                f" for value {value} ({type(value)})"
            )
            return False
    return True


def validate_graph_dataclasses_and_nodes(graph: Graph) -> None:
    log.debug("Validating attribute types of all graph dataclasses")
    node_chksums = {}
    invalid_nodes = []
    for node in graph.nodes:
        if isinstance(node, BaseResource):
            try:
                if not validate_dataclass(node):
                    invalid_nodes.append(node)
            except Exception:
                log.error(f"Failed to validate dataclass {node.kdname}")
            if node.chksum not in node_chksums:
                node_chksums[node.chksum] = node
            else:
                log.error(
                    f"Duplicate checksum {node.chksum} for node {node.kdname} in graph!\n"
                    f"Previous node: {get_resource_attributes(node_chksums[node.chksum])}\n"
                    f"Previous predecessor nodes: {[n.kdname for n in node_chksums[node.chksum].predecessors()]}\n"
                    f"Previous successor nodes: {[n.kdname for n in node_chksums[node.chksum].successors()]}\n"
                    f"New node: {get_resource_attributes(node)}\n"
                    f"New predecessor nodes: {[n.kdname for n in node.predecessors()]}\n"
                    f"New successor nodes: {[n.kdname for n in node.successors()]}\n"
                )
    for node in invalid_nodes:
        log.error(f"Removing invalid node {node.kdname} from graph")
        graph.remove_node(node)


def update_graph_ref(graph: Graph) -> None:
    for node in graph.nodes:
        if isinstance(node, BaseResource):
            node._graph = graph


def sanitize(graph: Graph, maybe_root: Optional[GraphRoot] = None) -> None:
    log.debug("Sanitizing Graph")
    plugin_roots = {}
    graph_roots = []

    root: Optional[BaseResource] = maybe_root

    if root is None and isinstance(getattr(graph, "root", None), BaseResource):
        root = graph.root

    if root is None:
        log.debug("No graph root found - unable to sanitize")
        return

    for node in graph.successors(root):
        if isinstance(node, Cloud):
            log.debug(f"Found Plugin Root {node.id}")
            plugin_roots[node.id] = node
        elif isinstance(node, GraphRoot):
            log.debug(f"Found Graph Root {node.id}")
            graph_roots.append(node)
        else:
            log.debug(f"Found unknown node {node.id} of type {node.kind}")

    for node in list(graph.nodes):
        if (
            isinstance(node, Cloud)
            and node not in plugin_roots.values()
            and len(list(graph.predecessors(node))) == 0
            and node.id in plugin_roots
        ):
            log.debug(
                f"Found unconnected plugin root {node.id} with same id as existing plugin root"
                " - moving child nodes to existing plugin root and removing unconnected plugin root"
            )
            for node_child in list(graph.successors(node)):
                graph.add_edge(plugin_roots[node.id], node_child)
                graph.remove_edge(node, node_child)
            graph.remove_node(node)

    if len(graph_roots) > 0:
        for graph_root in graph_roots:
            log.debug(f"Moving children of graph root {graph_root.id}")
            for node in list(graph.successors(graph_root)):
                if isinstance(node, Cloud):
                    if node.id in plugin_roots:
                        log.debug(
                            f"Found existing plugin root {node.id}" " - attaching children and removing plugin root"
                        )
                        for plugin_root_child in list(graph.successors(node)):
                            log.debug(
                                f"Found node {plugin_root_child.id} of type "
                                f"{plugin_root_child.kind}"
                                " - attaching to existing plugin root"
                            )
                            graph.add_edge(plugin_roots[node.id], plugin_root_child)
                            graph.remove_edge(node, plugin_root_child)
                        graph.remove_node(node)
                    else:
                        log.debug(f"Found new plugin root {node.id}" " - attaching to top level root")
                        graph.add_edge(root, node)
                        graph.remove_edge(graph_root, node)
                else:
                    log.debug(f"Found unknown node {node.id} of type {node.kind}" " - attaching to top level root")
                    graph.add_edge(root, node)
                    graph.remove_edge(graph_root, node)
            log.debug(f"Removing graph root {graph_root.id}")
            graph.remove_node(graph_root)
    graph.resolve_deferred_connections()
    update_graph_ref(graph)
    validate_graph_dataclasses_and_nodes(graph)


class GraphMergeKind(Enum):
    cloud = "cloud"
    account = "account"


class GraphExportIterator:
    def __init__(
        self,
        graph: Graph,
        delete_tempfile: bool = True,
        tempdir: Optional[str] = None,
        graph_merge_kind: GraphMergeKind = GraphMergeKind.cloud,
    ):
        self.graph = graph
        ts = datetime.now().strftime("%Y-%m-%d-%H-%M")
        self.tempfile = tempfile.NamedTemporaryFile(
            prefix=f"resoto-graph-{ts}-",
            suffix=".ndjson",
            delete=delete_tempfile,
            dir=tempdir,
        )
        if not delete_tempfile:
            log.info(f"Writing graph json to file {self.tempfile.name}")

        if not isinstance(graph_merge_kind, GraphMergeKind):
            log.error(f"Graph merge kind is wrong type {type(graph_merge_kind)}")
            graph_merge_kind = GraphMergeKind.cloud

        if graph_merge_kind == GraphMergeKind.cloud:
            self.graph_merge_kind: Type[Any] = BaseCloud
        elif graph_merge_kind == GraphMergeKind.account:
            self.graph_merge_kind = BaseAccount
        else:
            log.error(f"Graph merge kind has unknown value {graph_merge_kind} - defaulting to 'cloud'")
            self.graph_merge_kind = BaseCloud

        self.graph_exported = False
        self.found_replace_node = False
        self.export_lock = threading.Lock()
        self.total_lines = 0
        self.number_of_nodes = int(graph.number_of_nodes())
        self.number_of_edges = int(graph.number_of_edges())
        self.number_of_deferred_edges = len(graph.deferred_edges)

    def __del__(self) -> None:
        try:
            self.tempfile.close()
        except Exception:
            pass

    def __iter__(self) -> Iterator[bytes]:
        if not self.graph_exported:
            self.export_graph()
        start_time = time()
        last_sent = time()
        lines_sent = 0
        percent = 0
        report_every = round(self.total_lines / 10)

        self.tempfile.seek(0)
        while line := self.tempfile.readline():
            lines_sent += 1
            if report_every > 0 and lines_sent > 0 and lines_sent % report_every == 0:
                percent = round(lines_sent / self.total_lines * 100)
                elapsed = time() - last_sent
                log.debug(f"Sent {lines_sent}/{self.total_lines} nodes and edges ({percent}%) - {elapsed:.4f}s")
                last_sent = time()
            yield line
        elapsed = time() - start_time
        log.info(
            f"Sent {lines_sent}/{self.total_lines},"
            f" {self.number_of_nodes} nodes, {self.number_of_edges} edges"
            f" and {self.number_of_deferred_edges} deferred edges"
            f" in {elapsed:.4f}s"
        )

    def export_graph(self) -> None:
        with self.export_lock:
            start_time = time()
            for node in self.graph.nodes:
                node_dict = node_to_dict(node)
                if isinstance(node, self.graph_merge_kind):
                    log.debug(f"Replacing sub graph below {node.kdname}")
                    if "metadata" not in node_dict or not isinstance(node_dict["metadata"], dict):
                        node_dict["metadata"] = {}
                    node_dict["metadata"]["replace"] = True
                    self.found_replace_node = True
                if isinstance(node, BaseAccount):
                    log.debug(f"Setting export time on {node.kdname}")
                    if "metadata" not in node_dict or not isinstance(node_dict["metadata"], dict):
                        node_dict["metadata"] = {}
                    node_dict["metadata"]["exported_at"] = utc_str()
                node_json = to_json_str(node_dict) + "\n"
                self.tempfile.write(node_json.encode())
                self.total_lines += 1
            elapsed_nodes = time() - start_time
            log.debug(f"Exported {self.number_of_nodes} nodes in {elapsed_nodes:.4f}s")
            if not self.found_replace_node:
                log.warning(f"No nodes of kind {self.graph_merge_kind.kind} found in graph")
            start_time = time()
            for edge in self.graph.edges:
                from_node = edge[0]
                to_node = edge[1]
                if not isinstance(from_node, BaseResource) or not isinstance(to_node, BaseResource):
                    log.error(f"One of {from_node} and {to_node} is no base resource")
                    continue
                edge_dict = {"from": from_node.chksum, "to": to_node.chksum}
                if len(edge) == 3:
                    key = edge[2]
                    if isinstance(key, EdgeKey) and key.edge_type != EdgeType.default:
                        edge_dict["edge_type"] = key.edge_type.value
                edge_json = json.dumps(edge_dict) + "\n"
                self.tempfile.write(edge_json.encode())
                self.total_lines += 1
            for from_selector, to_selector, edge_type in self.graph.deferred_edges:
                deferred_edge_dict: Json = {}
                if isinstance(from_selector, ByNodeId):
                    deferred_edge_dict["from_selector"] = {"node_id": from_selector.value}
                else:
                    deferred_edge_dict["from_selector"] = {"search_criteria": from_selector.query}
                if isinstance(to_selector, ByNodeId):
                    deferred_edge_dict["to_selector"] = {"node_id": to_selector.value}
                else:
                    deferred_edge_dict["to_selector"] = {"search_criteria": to_selector.query}
                deferred_edge_dict["edge_type"] = edge_type.value
                deferred_edge_json = json.dumps(deferred_edge_dict) + "\n"
                self.tempfile.write(deferred_edge_json.encode())
                self.total_lines += 1
            elapsed_edges = time() - start_time
            log.debug(f"Exported {self.number_of_edges} edges in {elapsed_edges:.4f}s")
            elapsed = elapsed_nodes + elapsed_edges
            log.info(f"Exported {self.total_lines} nodes and edges in {elapsed:.4f}s")
            self.graph_exported = True
            # The graph uses attributes with @cached_property.
            # Those attributes reference the graph and increment the refcount.
            # See: https://github.com/networkx/networkx/issues/6235
            unset_cached_properties(self.graph)
            del self.graph
            self.tempfile.seek(0)



import json
import sys
from datetime import timedelta, datetime, date
from typing import TypeVar, Any, Type, Optional, Union, List, get_args, Literal, get_origin, Callable, Iterable, Set

from dateutil.parser import isoparse

if sys.version_info >= (3, 10):
    from types import UnionType, NoneType
else:
    UnionType = Union
    NoneType = type(None)

import attrs
import cattrs
from cattrs import override
from cattrs.gen import make_dict_unstructure_fn

from resotolib.durations import parse_duration, duration_str
from resotolib.logger import log
from resotolib.types import Json, JsonElement
from resotolib.utils import utc_str

AnyT = TypeVar("AnyT")

# the global converter instance
__converter = cattrs.Converter()

# ignore all private attributes
__converter.register_unstructure_hook_factory(
    attrs.has,
    lambda cls: make_dict_unstructure_fn(
        cls,
        __converter,
        _cattrs_omit_if_default=False,
        _cattrs_use_linecache=True,
        _cattrs_use_alias=False,
        _cattrs_include_init_false=False,
        **{a.name: override(omit=True) for a in attrs.fields(cls) if a.name.startswith("_")},
    ),
)


# work around until this is solved: https://github.com/python-attrs/cattrs/issues/278
def is_primitive_or_primitive_union(t: Any) -> bool:
    if t in (str, bytes, int, float, bool, NoneType):
        return True
    origin = get_origin(t)
    if origin is Literal:
        return True
    if (base := cattrs._compat.get_newtype_base(t)) is not None:
        return is_primitive_or_primitive_union(base)
    if origin in (UnionType, Union):
        return all(is_primitive_or_primitive_union(ty) for ty in get_args(t))
    return False


# allow timedelta either as number of seconds or as duration string
def timedelta_from_json(js: Any) -> timedelta:
    if isinstance(js, str):
        return parse_duration(js)
    elif isinstance(js, (int, float)):
        return timedelta(seconds=js)
    else:
        raise ValueError(f"Cannot convert {js} to timedelta")


__converter.register_structure_hook_func(is_primitive_or_primitive_union, lambda v, ty: v)


def register_json(
    cls: Type[AnyT],
    to_json_fn: Optional[Callable[[AnyT], JsonElement]] = None,
    from_json_fn: Optional[Callable[[Any], AnyT]] = None,
) -> None:
    """
    Register a json marshaller/unmarshaller for the given class.
    :param cls: the class to register
    :param to_json_fn: the function to convert the class to json
    :param from_json_fn: the function to convert json to the class
    """
    log.trace("Register json structure hooks for class %s", cls.__name__)
    if from_json_fn is not None:
        __converter.register_structure_hook(cls, lambda obj, _: from_json_fn(obj))
    if to_json_fn is not None:
        __converter.register_unstructure_hook(cls, to_json_fn)


# Register some default types not covered in cattrs
register_json(datetime, utc_str, isoparse)
register_json(date, lambda obj: obj.isoformat(), date.fromisoformat)
register_json(timedelta, duration_str, timedelta_from_json)


def to_json_str(node: Any, strip_attr: Union[None, str, Iterable[str]] = None, strip_nulls: bool = False) -> str:
    try:
        return json.dumps(to_json(node, strip_attr, strip_nulls))
    except Exception as e:
        log.debug(f"Can not serialize object {node} to json. Error: {e}")
        raise


def to_json(
    node: Any,
    strip_attr: Union[None, str, Iterable[str]] = None,
    strip_nulls: bool = False,
    keep_untouched: Optional[Set[str]] = None,
) -> Json:
    """
    Use this method, if the given node is known as complex object,
    so the result will be a json object.
    """

    def walk_js_object(js: Json, filter_fn: Optional[Callable[[str, Any], bool]] = None) -> Json:
        result: Json = {}
        for k, v in js.items():
            if keep_untouched and k in keep_untouched:
                result[k] = v
                continue
            if filter_fn and not filter_fn(k, v):
                continue
            if isinstance(v, dict):
                v = walk_js_object(v, filter_fn)
            elif isinstance(v, (list, tuple)):
                v = [walk_js_object(e, filter_fn) if isinstance(e, dict) else e for e in v]
            result[k] = v
        return result

    unstructured: Json = __converter.unstructure(node)
    if strip_attr:
        remove_keys = {strip_attr} if isinstance(strip_attr, str) else set(strip_attr)
        unstructured = walk_js_object(unstructured, lambda k, v: k not in remove_keys)

    if strip_nulls:
        unstructured = walk_js_object(unstructured, lambda k, v: v is not None)

    return unstructured


def from_json(js: JsonElement, clazz: Type[AnyT]) -> AnyT:
    """
    Loads a json object into a python object.
    :param js: the json object to load.
    :param clazz: the type of the python object.
    :return: the loaded python object.
    """
    try:
        return __converter.structure(js, clazz)
    except Exception as e:
        log.debug(f"Can not deserialize json into class {clazz.__name__}: {js}. Error: {e}")
        raise


def value_in_path(element: JsonElement, path_or_name: Union[List[str], str]) -> Optional[Any]:
    """
    Access a value in a json object by a defined path.
    {"a": {"b": {"c": 1}}} -> value_in_path({"a": {"b": {"c": 1}}}, ["a", "b", "c"]) -> 1
    The path can be defined as a list of strings or as a string with dots as separator.
    """
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path)

    def at_idx(current: JsonElement, idx: int) -> Optional[Any]:
        if at == idx:
            return current
        elif current is None or not isinstance(current, dict) or path[idx] not in current:
            return None
        else:
            return at_idx(current[path[idx]], idx + 1)

    return at_idx(element, 0)


def set_value_in_path(element: JsonElement, path_or_name: Union[List[str], str], js: Optional[Json] = None) -> Json:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path) - 1

    def at_idx(current: Json, idx: int) -> None:
        if at == idx:
            current[path[-1]] = element
        else:
            value = current.get(path[idx])
            if not isinstance(value, dict):
                value = {}
                current[path[idx]] = value
            at_idx(value, idx + 1)

    js = js if js is not None else {}
    at_idx(js, 0)
    return js


def is_empty(js: JsonElement) -> bool:
    if js is None:
        return True
    elif isinstance(js, dict):
        return all(is_empty(v) for v in js.values())
    elif isinstance(js, list):
        return all(is_empty(v) for v in js)
    else:
        return False


def sort_json(js_object: Json, *, sort_list: bool = False) -> Json:
    def walk(js: JsonElement) -> JsonElement:
        if isinstance(js, dict):
            # sort by
            return {k: walk(v) for k, v in sorted(js.items())}
        elif isinstance(js, list):
            gen = (walk(v) for v in js)
            return list(sorted(gen, key=lambda x: 1 if isinstance(x, (dict, list)) else x)) if sort_list else list(gen)  # type: ignore # noqa
        else:
            return js

    return walk(js_object)  # type: ignore



import os
import jwt
import base64
import hashlib
import time

from cryptography.hazmat.primitives import serialization
from jwt.utils import base64url_encode

from resotolib.args import ArgumentParser
from resotolib.x509 import x5t_s256, x5t
from typing import Any, Optional, Tuple, Dict, Mapping, Union, cast
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from cryptography.x509.base import Certificate

from resotolib.types import Json


def key_from_psk(psk: str, salt: Optional[bytes] = None) -> Tuple[bytes, bytes]:
    """Derive a 256 bit key from a passphrase/pre-shared-key.
    A salt can be optionally provided. If not one will be generated.
    Returns both the key and the salt.
    """
    if salt is None:
        salt = os.urandom(16)
    key = hashlib.pbkdf2_hmac("sha256", psk.encode(), salt, 100000)
    return key, salt


def encode_jwt(
    payload: Dict[str, Any],
    key: Union[str, RSAPrivateKey],
    headers: Optional[Dict[str, str]] = None,
    expire_in: int = 300,
    cert: Optional[Certificate] = None,
) -> str:
    """Encodes a payload into a JWT either using a pre-shared-key or an RSA private key."""
    payload = dict(payload)
    if headers is None:
        headers = {}
    headers.update({"typ": "JWT"})
    if expire_in > 0 and "exp" not in payload:
        payload.update({"exp": int(time.time()) + expire_in})

    if isinstance(key, RSAPrivateKey):
        return encode_jwt_pki(payload, key, headers, cert)
    else:
        return encode_jwt_psk(payload, key, headers)


def encode_jwt_pki(
    payload: Dict[str, Any],
    private_key: RSAPrivateKey,
    headers: Dict[str, str],
    cert: Optional[Certificate] = None,
) -> str:
    """Encodes a payload into a JWT either using an RSA private key."""
    headers.update({"alg": "RS256"})
    if cert is not None:
        headers.update({"x5t#S256": x5t_s256(cert)})
    return jwt.encode(payload, private_key, algorithm="RS256", headers=headers)


def encode_jwt_psk(
    payload: Dict[str, Any],
    psk: str,
    headers: Dict[str, str],
) -> str:
    """Encodes a payload into a JWT and signs using a key derived from a pre-shared-key.
    Stores the key's salt in the JWT headers.
    """
    key, salt = key_from_psk(psk)
    salt_encoded = base64.standard_b64encode(salt).decode("utf-8")
    headers.update({"alg": "HS256", "salt": salt_encoded})
    return jwt.encode(payload, key, algorithm="HS256", headers=headers)


def decode_jwt(
    encoded_jwt: str, psk_or_cert: Union[str, Certificate, RSAPublicKey], options: Optional[Dict[str, Any]] = None
) -> Json:
    """Decode a JWT using a key derived from a pre-shared-key and a salt stored
    in the JWT headers or an RSA public key.
    """
    alg = jwt.get_unverified_header(encoded_jwt).get("alg")
    if alg == "RS256":
        assert isinstance(psk_or_cert, (Certificate, RSAPublicKey))
        return decode_jwt_pki(encoded_jwt, psk_or_cert, options)
    elif alg == "HS256":
        assert isinstance(psk_or_cert, str)
        return decode_jwt_psk(encoded_jwt, psk_or_cert, options)
    else:
        raise ValueError(f"Unsupported JWT algorithm: {alg}")


def decode_jwt_psk(encoded_jwt: str, psk: str, options: Optional[Dict[str, Any]] = None) -> Json:
    """Decode a JWT using a key derived from a pre-shared-key and a salt stored
    in the JWT headers.
    """
    salt_encoded = jwt.get_unverified_header(encoded_jwt).get("salt")
    salt = base64.standard_b64decode(salt_encoded) if salt_encoded else None
    key, _ = key_from_psk(psk, salt)
    return jwt.decode(encoded_jwt, key, algorithms=["HS256"], options=options)  # type: ignore


def decode_jwt_pki(
    encoded_jwt: str, public_key: Union[Certificate, RSAPublicKey], options: Optional[Dict[str, Any]] = None
) -> Json:
    """Decode a JWT using an RSA public key."""
    if isinstance(public_key, Certificate):
        public_key = cast(RSAPublicKey, public_key.public_key())
        assert isinstance(public_key, RSAPublicKey)
    return jwt.decode(encoded_jwt, public_key, algorithms=["RS256"], options=options)  # type: ignore


def encode_jwt_to_headers(
    http_headers: Dict[str, str],
    payload: Dict[str, Any],
    key: Union[str, RSAPrivateKey],
    scheme: str = "Bearer",
    headers: Optional[Dict[str, str]] = None,
    expire_in: int = 300,
    cert: Optional[Certificate] = None,
) -> Dict[str, str]:
    """Takes a payload and psk turns them into a JWT and adds that to a http headers
    dictionary. Also returns that dict.
    """
    http_headers.update({"Authorization": f"{scheme} {encode_jwt(payload, key, headers, expire_in, cert)}"})
    return http_headers


def decode_jwt_from_headers(
    http_headers: Mapping[str, str],
    psk_or_cert: Union[str, Certificate, RSAPublicKey],
    scheme: str = "Bearer",
    options: Optional[Dict[str, Any]] = None,
) -> Optional[Dict[str, str]]:
    """Retrieves the Authorization header from a http headers dictionary and
    passes it to `decode_jwt_from_header_value()` to return the decoded payload.
    """
    authorization_header = {str(k).capitalize(): v for k, v in http_headers.items()}.get("Authorization")
    if authorization_header is None:
        return None
    return decode_jwt_from_header_value(authorization_header, psk_or_cert, scheme, options)


def decode_jwt_from_header_value(
    authorization_header: str,
    psk_or_cert: Union[str, Certificate, RSAPublicKey],
    scheme: str = "Bearer",
    options: Optional[Dict[str, Any]] = None,
) -> Optional[Dict[str, str]]:
    """Decodes a JWT payload from a http Authorization header value."""
    if (
        len(authorization_header) <= len(scheme) + 1
        or str(authorization_header[0 : len(scheme)]).lower() != scheme.lower()
        or authorization_header[len(scheme) : len(scheme) + 1] != " "
    ):
        return None
    encoded_jwt = authorization_header[len(scheme) + 1 :]
    return decode_jwt(encoded_jwt, psk_or_cert, options)


def create_jwk_dict(cert: Certificate) -> Json:
    pub_key = cert.public_key()
    if not isinstance(pub_key, RSAPublicKey):
        raise ValueError(f"Unsupported public key type: {type(pub_key)}")
    pem_data = cert.public_bytes(serialization.Encoding.PEM)
    pem_contents = pem_data.split(b"\n")[1:-2]  # Remove header and footer lines
    x5t_256 = x5t_s256(cert)
    pub_num = pub_key.public_numbers()
    return {
        "kty": "RSA",
        "alg": cert.signature_algorithm_oid._name,
        "n": base64url_encode(pub_num.n.to_bytes((pub_num.n.bit_length() + 7) // 8, "big")).decode("utf-8"),
        "e": base64url_encode(pub_num.e.to_bytes((pub_num.e.bit_length() + 7) // 8, "big")).decode("utf-8"),
        "use": "sig",
        "kid": x5t_256,
        "x5t": x5t(cert),
        "x5t#S256": x5t_256,
        "x5c": [b"".join(pem_contents).decode("utf-8")],
    }


def add_args(arg_parser: ArgumentParser) -> None:
    arg_parser.add_argument(
        "--psk",
        help="Pre-shared key",
        type=lambda x: x if len(x) > 0 else None,
        default=None,
        dest="psk",
    )



from typing import Any, Dict, List, Type, Optional, Callable

from resotolib.types import DecoratedFn

definitions = "__task_definitions"


class CommandDefinition:
    """
    A final decorator is called by the interpreter once after the class is defined.
    (Not the case for decorators with attributes)
    This decorator will maintain all decorated methods on the class side.
    Use task_definitions(clazz) to get all task definitions of a specific class.
    """

    def __init__(
        self,
        fn: DecoratedFn,
        name: str,
        info: str,
        args_description: Dict[str, str],
        description: str,
        expect_node_result: bool = False,
        expect_resource: bool = False,
        allowed_on_kind: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        self.fn = fn
        self.name = name
        self.info = info
        self.args_description = args_description
        self.description = description
        self.expect_resource = expect_resource
        self.expect_node_result = expect_node_result
        self.allowed_on_kind = allowed_on_kind
        self.filter = filter or {}

    def __set_name__(self, owner: str, name: str) -> None:
        # store this definition on the owners side (class)
        if getattr(owner, definitions, None) is None:
            setattr(owner, definitions, [])
        getattr(owner, definitions).append(self)
        # make the function available at the call side (class)
        setattr(owner, name, self.fn)

    def __call__(self) -> Callable[..., Any]:
        return self.fn


# noinspection PyPep8Naming
class execute_command:  # noqa: N801
    """
    In case you want to expose a method as worker task to the core, you can use this decorator.
    The definition is used to register a custom command at the core, when the worker connects.
    The custom command can be executed with the provided name.
    Note: any existing command or alias takes precedence over the custom command.

    The expected signature of the function is this:

    def some_name(self, config: Config, js: Json, args: List[str]) -> JsonElement:

    Notes on the signature:
    - the json is the node section of the core task data.
    - the args contains the list of arguments provided by the user.
    - the function returns json. None is a valid json element.

    :param name: the name of the command
    :param info: a short description of the command
    :param args_description: a dictionary of argument names and their description
    :param description: a longer description of the command
    :param expect_node_result: if true, the returned json is a resource and the current node in the db will be updated.
    :param allowed_on_kind: if set, the command will only be executed on resources of this kind
    :param filter: filter attributes as selector for this command.
           Multiple workers might register the same command on the same resource with different filters.
           Based on the resource the task is delegated to the worker with the matching filter.
    :return: the decorated function that will be registered as worker task.
    """

    def __init__(
        self,
        name: str,
        info: str,
        args_description: Dict[str, str],
        description: str,
        expect_node_result: bool = False,
        allowed_on_kind: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        self.name = name
        self.info = info
        self.args_description = args_description
        self.description = description
        self.expect_node_result = expect_node_result
        self.allowed_on_kind = allowed_on_kind
        self.filter = filter or {}

    def __call__(self, fn: DecoratedFn) -> CommandDefinition:
        return CommandDefinition(
            fn,
            name=self.name,
            info=self.info,
            args_description=self.args_description,
            description=self.description,
            expect_node_result=self.expect_node_result,
            expect_resource=False,
            allowed_on_kind=self.allowed_on_kind,
            filter=self.filter,
        )


# noinspection PyPep8Naming
class execute_command_on_resource:  # noqa: N801
    """
    See execute_command.

    This decorator will convert the incoming message to a resource object
    and a resulting object back to the json representation.

    The expected signature of the function is this:

    def some_name(
        self, config: Config, resource: Optional[BaseResource], args: List[str]
    ) -> Union[JsonElement, BaseResource]:

    Notes on the signature:
    - the resource is marked as optional. In case the command is executed in source position, no resource is provided.
    - the args contains the list of arguments provided by the user.
    - the function can decide to return either a json element or a resource. None is a valid json element.

    :param name: the name of the command
    :param info: a short description of the command
    :param args_description: a dictionary of argument names and their description
    :param description: a longer description of the command
    :param expect_node_result: if true, the returned resource will update the current node in the database.
    :param allowed_on_kind: if set, the command will only be executed on resources of this kind
    :param filter: filter attributes as selector for this command.
           Multiple workers might register the same command on the same resource with different filters.
           Based on the resource the task is delegated to the worker with the matching filter.
    :return: the decorated function that will be registered as worker task.
    """

    def __init__(
        self,
        name: str,
        info: str,
        args_description: Dict[str, str],
        description: str,
        expect_node_result: bool = False,
        allowed_on_kind: Optional[str] = None,
        filter: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        self.name = name
        self.info = info
        self.args_description = args_description
        self.description = description
        self.expect_node_result = expect_node_result
        self.allowed_on_kind = allowed_on_kind
        self.filter = filter or {}

    def __call__(self, fn: DecoratedFn) -> CommandDefinition:
        return CommandDefinition(
            fn,
            name=self.name,
            info=self.info,
            args_description=self.args_description,
            description=self.description,
            expect_node_result=self.expect_node_result,
            expect_resource=True,
            allowed_on_kind=self.allowed_on_kind,
            filter=self.filter,
        )


def command_definitions(clazz: Type[Any]) -> List[CommandDefinition]:
    return getattr(clazz, definitions, [])



import threading
import time
import json
from urllib.parse import urlencode

from websocket import WebSocketApp, WebSocket  # type: ignore

from resotolib.logger import log
from resotolib.event import EventType, remove_event_listener, add_event_listener, Event
from resotolib.args import ArgumentParser
from resotolib.jwt import encode_jwt_to_headers
from resotolib.core.ca import TLSData
from typing import Callable, Dict, Optional, Set, Any

from resotolib.types import Json


class CoreEvents(threading.Thread):
    def __init__(
        self,
        resotocore_ws_uri: str,
        events: Optional[Set[str]] = None,
        message_processor: Optional[Callable[[Json], None]] = None,
        tls_data: Optional[TLSData] = None,
    ) -> None:
        super().__init__()
        self.ws_uri = f"{resotocore_ws_uri}/events"
        if events:
            query_string = urlencode({"show": ",".join(events)})
            self.ws_uri += f"?{query_string}"
        self.message_processor = message_processor
        self.tls_data = tls_data
        self.ws: Optional[WebSocketApp] = None
        self.shutdown_event = threading.Event()
        self.__connected = False

    def connected(self) -> bool:
        return self.__connected

    def __del__(self) -> None:
        remove_event_listener(EventType.SHUTDOWN, self.shutdown)

    def run(self) -> None:
        self.name = "eventbus-listener"
        add_event_listener(EventType.SHUTDOWN, self.shutdown)
        while not self.shutdown_event.is_set():
            log.debug("Connecting to resotocore event bus")
            try:
                self.connect()
            except Exception as e:
                log.error(e)
            time.sleep(1)

    def connect(self) -> None:
        log.debug(f"Connecting to {self.ws_uri}")
        headers: Dict[str, str] = {}
        if getattr(ArgumentParser.args, "psk", None):
            encode_jwt_to_headers(headers, {}, ArgumentParser.args.psk)
        self.ws = WebSocketApp(
            self.ws_uri,
            header=headers,
            on_open=self.on_open,
            on_message=self.on_message,
            on_error=self.on_error,
            on_close=self.on_close,
            on_ping=self.on_ping,
            on_pong=self.on_pong,
        )
        sslopt: Dict[Any, Any] = {}
        if self.tls_data:
            sslopt = {"ca_certs": self.tls_data.ca_cert_path}
        self.ws.run_forever(sslopt=sslopt, ping_interval=20, ping_timeout=10, ping_payload="ping")

    def shutdown(self, _: Optional[Event] = None) -> None:
        log.debug("Received shutdown event - shutting down resotocore event bus listener")
        self.shutdown_event.set()
        if self.ws:
            self.ws.close()

    def on_message(self, _: WebSocket, message: str) -> None:
        try:
            json_message: Json = json.loads(message)
        except json.JSONDecodeError:
            log.exception(f"Unable to decode received message {message}")
            return
        log.debug(f"Received event: {json_message}")
        if self.message_processor is not None and callable(self.message_processor):
            try:
                self.message_processor(json_message)
            except Exception:
                log.exception(f"Something went wrong while processing {message}")

    def on_error(self, _: WebSocket, e: Exception) -> None:
        log.debug(f"Event bus error: {e!r}")

    def on_close(self, _: WebSocket, close_status_code: int, close_msg: str) -> None:
        self.__connected = False
        log.debug("Disconnected from resotocore event bus")

    def on_open(self, _: WebSocket) -> None:
        self.__connected = True
        log.debug("Connected to resotocore event bus")

    def on_ping(self, _: WebSocket, message: str) -> None:
        log.debug("Ping from resotocore event bus")

    def on_pong(self, _: WebSocket, message: str) -> None:
        log.debug("Pong from resotocore event bus")



import hashlib
import os
import random
import re
import socket
import string
import sys
import time
from argparse import ArgumentParser
from contextlib import closing
from copy import deepcopy
from datetime import date, datetime, timezone, timedelta
from functools import wraps, cached_property
from tarfile import TarFile, TarInfo
from typing import (
    Dict,
    List,
    Tuple,
    Optional,
    NoReturn,
    Any,
    Mapping,
    Union,
    Callable,
    cast,
    Iterator,
    TypeVar,
    Sequence,
)
from zoneinfo import ZoneInfo

import select
from frozendict import frozendict
from tzlocal import get_localzone_name

from resotolib.logger import log
from resotolib.types import DecoratedFn, JsonElement

T = TypeVar("T")
UTC_Date_Format = "%Y-%m-%dT%H:%M:%SZ"


def rnd_str(str_len: int = 10) -> str:
    return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(str_len))


def utc() -> datetime:
    return datetime.now(timezone.utc)


def utc_str(dto: Optional[datetime] = None) -> str:
    dt = dto if dto is not None else utc()
    if dt.tzinfo is not None and dt.tzname() != "UTC":
        offset = dt.tzinfo.utcoffset(dt)
        if offset is not None and offset.total_seconds() != 0:
            dt = (dt - offset).replace(tzinfo=timezone.utc)
    return dt.strftime(UTC_Date_Format)


def parse_utc(date_string: str) -> datetime:
    dt = datetime.fromisoformat(date_string)
    if (
        not dt.tzinfo
        or dt.tzinfo.utcoffset(None) is None
        or dt.tzinfo.utcoffset(None).total_seconds() != 0  # type: ignore
    ):
        dt = dt.astimezone(timezone.utc)
    return dt


def make_valid_timestamp(timestamp: datetime) -> Optional[datetime]:
    if not isinstance(timestamp, datetime) and isinstance(timestamp, date):
        timestamp = datetime.combine(timestamp, datetime.min.time()).replace(tzinfo=timezone.utc)
    elif isinstance(timestamp, datetime) and timestamp.tzinfo is None:
        timestamp = timestamp.replace(tzinfo=timezone.utc)
    elif isinstance(timestamp, datetime):
        pass
    else:
        timestamp = None
    return timestamp


def get_local_tzinfo() -> ZoneInfo:
    zone_name = get_localzone_name()
    if zone_name is None:
        zone_name = "Etc/UTC"
    return ZoneInfo(zone_name)


def chunks(items: List[T], n: int) -> Iterator[List[T]]:
    """Split a list of items into multiple lists of size n and yield each chunk"""
    for s in range(0, len(items), n):
        e = s + n
        yield items[s:e]


def unset_cached_properties(obj: Any) -> None:
    """
    Reset all cached properties of an object.
    Successive calls to the property will recompute the value.
    :param obj: the object with cached properties.
    """
    cls = obj.__class__
    for a in dir(obj):
        attr_a = getattr(cls, a, cls)
        if isinstance(attr_a, cached_property):
            obj.__dict__.pop(attr_a.attrname, None)


# via https://stackoverflow.com/a/1094933
def iec_size_format(byte_size: float) -> str:
    for unit in ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB"]:
        if abs(byte_size) < 1024.0:
            return f"{byte_size:.2f} {unit}"
        byte_size /= 1024.0
    return f"{byte_size:.2f} YiB"


# via https://stackoverflow.com/a/44873382
def sha256sum(filename: str, buffer_size: int = 128 * 1024) -> str:
    h = hashlib.sha256()
    buffer = bytearray(buffer_size)
    buffer_view = memoryview(buffer)
    with open(filename, "rb", buffering=0) as f:
        while True:
            n = f.readinto(buffer_view)
            if not n:
                break
            h.update(buffer_view[:n])
    return h.hexdigest()


def log_runtime(f: DecoratedFn) -> DecoratedFn:
    @wraps(f)
    def timer(*args: Any, **kwargs: Any) -> Any:
        start = time.time()
        ret = f(*args, **kwargs)
        runtime = time.time() - start
        args_str = ", ".join([repr(arg) for arg in args])
        kwargs_str = ", ".join([f"{k}={repr(v)}" for k, v in kwargs.items()])
        if len(args) > 0 and len(kwargs) > 0:
            args_str += ", "
        log.debug(f"Runtime of {f.__name__}({args_str}{kwargs_str}): {runtime:.3f} seconds")
        return ret

    return timer  # type: ignore


def except_log_and_pass(do_raise: Optional[Tuple[Any]] = None) -> Callable[..., Any]:
    do_raise_tuple = do_raise if do_raise is not None else ()

    def acallable(f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def catch_and_log(*args: Any, **kwargs: Any) -> Any:
            try:
                return f(*args, **kwargs)
            except do_raise_tuple:
                raise
            except Exception:
                args_str = ", ".join([repr(arg) for arg in args])
                kwargs_str = ", ".join([f"{k}={repr(v)}" for k, v in kwargs.items()])
                if len(args) > 0 and len(kwargs) > 0:
                    args_str += ", "
                log.exception(f"Caught exception in {f.__name__}({args_str}{kwargs_str})")

        return catch_and_log

    return acallable


resource_attributes_blacklist = ["event_log"]


def get_resource_attributes(
    resource: Any, exclude_private: bool = True, keep_data_structures: bool = False
) -> Dict[str, Any]:
    attributes = dict(resource.__dict__)
    attributes["kind"] = resource.kind

    for attr_name in dir(resource):
        if exclude_private and attr_name.startswith("_"):
            continue
        attr_type = getattr(type(resource), attr_name, None)
        if isinstance(attr_type, property):
            attributes[attr_name] = getattr(resource, attr_name, None)

    remove_keys = []
    add_keys = {}

    for key, value in attributes.items():
        if exclude_private and str(key).startswith("_") or str(key) in resource_attributes_blacklist:
            remove_keys.append(key)
        elif isinstance(value, (list, tuple, set)) and not keep_data_structures:
            remove_keys.append(key)
            for i, v in enumerate(value):
                if v is not None:
                    add_keys[key + "[" + str(i) + "]"] = v
        elif isinstance(value, dict) and not keep_data_structures:
            remove_keys.append(key)
            for k, v in value.items():
                if v is not None:
                    add_keys[key + "['" + k + "']"] = v
        elif isinstance(value, (date, datetime, timedelta)):
            attributes[key] = str(value)
        elif value is None:
            remove_keys.append(key)
        elif not isinstance(
            value,
            (
                str,
                int,
                float,
                complex,
                list,
                tuple,
                range,
                dict,
                set,
                frozenset,
                bool,
                bytes,
                bytearray,
                memoryview,
            ),
        ):
            remove_keys.append(key)

    for key in remove_keys:
        attributes.pop(key)
    attributes.update(add_keys)

    return attributes


def type_str(o: Any) -> str:
    cls = o.__class__
    module = str(cls.__module__)
    if module == "builtins":
        return str(cls.__qualname__)
    return module + "." + str(cls.__qualname__)


def get_local_ip_addresses(
    *,
    include_loopback: bool = True,
    san_ip_addresses: Optional[List[str]] = None,
    connect_to_ips: Optional[List[str]] = None,
) -> List[str]:
    ips = set()
    if connect_to_ips is None:
        connect_to_ips = ["8.8.8.8", "2001:4860:4860::8888"]
    if include_loopback:
        ips.add("127.0.0.1")
        ips.add("::1")

    try:
        local_address = socket.gethostbyname(socket.gethostname())
    except Exception:
        pass
    else:
        ips.add(local_address)

    for dst_ip in connect_to_ips:
        try:
            af_inet = socket.AF_INET6 if ":" in dst_ip else socket.AF_INET
            with socket.socket(af_inet, socket.SOCK_DGRAM) as s:
                s.connect((dst_ip, 53))
                local_address = s.getsockname()[0]
        except Exception:
            pass
        else:
            ips.add(local_address)

    if isinstance(san_ip_addresses, list):
        ips.update(san_ip_addresses)
    return list(ips)


def get_local_hostnames(
    *,
    include_loopback: bool = True,
    san_ip_addresses: Optional[List[str]] = None,
    san_dns_names: Optional[List[str]] = None,
    connect_to_ips: Optional[List[str]] = None,
) -> List[str]:
    hostnames = set()
    if include_loopback:
        hostnames.add("localhost")

    try:
        local_hostname = socket.gethostname()
    except Exception:
        pass
    else:
        hostnames.add(local_hostname)

    for ip in get_local_ip_addresses(
        include_loopback=include_loopback,
        connect_to_ips=connect_to_ips,
        san_ip_addresses=san_ip_addresses,
    ):
        try:
            hostname = socket.gethostbyaddr(ip)[0]
        except Exception:
            pass
        else:
            hostnames.add(hostname)

    if isinstance(san_dns_names, list):
        hostnames.update(san_dns_names)
    return list(hostnames)


def ordinal(num: int) -> str:
    suffix = "tsnrhtdd"[(num // 10 % 10 != 1) * (num % 10 < 4) * num % 10 :: 4]
    return f"{num}{suffix}"


def safe_members_in_tarfile(tarfile: TarFile) -> List[Any]:
    # via vmware/pyvcloud/vcd/utils.py
    def badpath(path: str, base: str) -> bool:
        # joinpath will ignore base if path is absolute
        return not os.path.realpath(os.path.abspath(os.path.join(base, path))).startswith(base)

    def badlink(info: TarInfo, base: str) -> bool:
        # Links are interpreted relative to the directory containing the link
        tip = os.path.realpath(os.path.abspath(os.path.join(base, os.path.dirname(info.name))))
        return badpath(info.linkname, base=tip)

    base = os.path.realpath(os.path.abspath((".")))
    basename = os.path.basename(tarfile.name)  # type: ignore
    result = []
    for tar_info in tarfile.getmembers():
        if badpath(tar_info.name, base):
            log.error(f"Error in {basename}, {tar_info.name} is blocked: illegal path")
        elif tar_info.issym() and badlink(tar_info, base):
            log.error(f"Error in {basename}, {tar_info.name} is blocked:" f" symlink to {tar_info.linkname}")
        elif tar_info.islnk() and badlink(tar_info, base):
            log.error(f"Error in {basename}, {tar_info.name} is blocked:" f" hard link to {tar_info.linkname}")
        else:
            result.append(tar_info)
    return result


def rrdata_as_dict(record_type: str, record_data: str) -> Dict[str, Any]:
    record_type = record_type.upper()
    rrdata: Dict[str, Any] = {}
    record_elements = []
    if record_type not in ("TXT"):
        record_data = " ".join(
            "".join([line.split(";")[0] for line in record_data.splitlines()]).replace("(", "").replace(")", "").split()
        )
    if record_type in ("SOA", "MX", "SRV", "CAA"):
        record_elements = record_data.split(" ")

    rrdata["record_value"] = record_data

    if record_type in ("A", "AAAA", "CNAME", "NS", "PTR"):
        pass
    elif record_type in ("TXT"):
        if record_data[0] == '"' and record_data[-1] == '"':
            record_data = record_data[1:-1]
        merge_pattern = '" "'
        merge_pattern_deletions = 0
        merge_pattern_offsets = [m.start() for m in re.finditer(merge_pattern, record_data)]
        for offset in merge_pattern_offsets:
            offset = offset - merge_pattern_deletions * len(merge_pattern)
            if record_data[offset - 1] == "\\":
                continue
            record_data = record_data[0:offset] + record_data[offset + len(merge_pattern) :]
            merge_pattern_deletions += 1

        rrdata["record_value"] = record_data
    elif record_type in ("SOA"):
        rrdata["record_value"] = record_data
        if len(record_elements) != 7:
            raise ValueError(f"Invalid SOA record {record_data}")
        rrdata["record_mname"] = record_elements[0]
        rrdata["record_rname"] = record_elements[1]
        rrdata["record_serial"] = int(record_elements[2])
        rrdata["record_refresh"] = int(record_elements[3])
        rrdata["record_retry"] = int(record_elements[4])
        rrdata["record_expire"] = int(record_elements[5])
        rrdata["record_minimum"] = int(record_elements[6])
    elif record_type in ("MX"):
        if len(record_elements) != 2:
            raise ValueError(f"Invalid MX record {record_data}")
        rrdata["record_priority"] = int(record_elements[0])
        rrdata["record_value"] = record_elements[1]
    elif record_type in ("SRV"):
        if len(record_elements) != 4:
            raise ValueError(f"Invalid SRV record {record_data}")
        rrdata["record_priority"] = int(record_elements[0])
        rrdata["record_weight"] = int(record_elements[1])
        rrdata["record_port"] = int(record_elements[2])
        rrdata["record_value"] = record_elements[3]
    elif record_type in ("CAA"):
        if len(record_elements) != 3:
            raise ValueError(f"Invalid CAA record {record_data}")
        rrdata["record_flags"] = int(record_elements[0])
        rrdata["record_tag"] = record_elements[1]
        rrdata["record_value"] = record_elements[2]

    return rrdata


class NoExitArgumentParser(ArgumentParser):
    def error(self, message: str) -> NoReturn:
        raise AttributeError(f"Could not parse arguments: {message}")

    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        msg = message if message else "unknown"
        raise AttributeError(f"Could not parse arguments: {msg}")


env_var_substitution_pattern = re.compile(r"\$\((\w+)\)")


def is_env_var_string(obj: Any) -> bool:
    is_string = isinstance(obj, str)
    if not is_string:
        return False
    has_env_var = re.search(env_var_substitution_pattern, obj) is not None
    return has_env_var


def replace_env_vars(elem: JsonElement, environment: Mapping[str, str], keep_unresolved: bool = True) -> JsonElement:
    # a special marker to avoid removing nulls
    class UnresolvedEnvVar:
        pass

    # no need to have many instances of this
    Unresolved = UnresolvedEnvVar()

    def replace_env_vars_helper(
        elem: JsonElement, environment: Mapping[str, str], keep_unresolved: bool, path: List[Union[str, int]]
    ) -> Union[JsonElement, UnresolvedEnvVar]:
        if isinstance(elem, dict):
            replaced_dict = {
                k: replace_env_vars_helper(v, environment, keep_unresolved, path + [k]) for k, v in elem.items()
            }
            without_unresolved_dict = {k: v for k, v in replaced_dict.items() if v is not Unresolved}
            return without_unresolved_dict
        elif isinstance(elem, list):
            replaced_list = [
                replace_env_vars_helper(v, environment, keep_unresolved, path + [i]) for i, v, in enumerate(elem)
            ]
            without_unresolved_list = [v for v in replaced_list if v is not Unresolved]
            return without_unresolved_list
        elif isinstance(elem, str):
            str_value = elem
            for match in re.finditer(env_var_substitution_pattern, elem):
                env_var_name = match.group(1)
                if env_var_found := environment.get(env_var_name):
                    str_value = str_value.replace(match.group(0), env_var_found)
                elif keep_unresolved:
                    pass
                else:
                    conf_path = ""

                    for idx, part in enumerate(path):
                        if idx == 0:
                            conf_path += str(part)
                        else:
                            conf_path += f"[{part}]" if isinstance(part, int) else f".{part}"
                    message = f"The environment variable `{env_var_name}` is not defined "
                    message += "in configuration at path {conf_path}. "
                    message += f"Please set the environment variable `{env_var_name}` or adjust the configuration. "
                    message += "You can also use the --override option to override the config value."
                    log.warning(f"Environment variable substitution failed: {message}")

                    return Unresolved

            return str_value
        elif isinstance(elem, UnresolvedEnvVar):
            # let's not leak that outside
            return None
        else:
            return elem

    replaced = replace_env_vars_helper(elem, environment, keep_unresolved, [])
    if isinstance(replaced, UnresolvedEnvVar):
        return None
    else:
        return replaced


def merge_json_elements(
    existing: JsonElement,
    update: JsonElement,
    merge_strategy: Callable[[JsonElement, JsonElement], JsonElement] = lambda existing_val, update_val: update_val,
) -> JsonElement:
    """
    Merges two JsonElements accorting to merge strategy.
    By default recursively traverses Dicts and prefers the new value
    """
    if isinstance(existing, dict) and isinstance(update, dict):
        output = deepcopy(existing)
        for update_key, update_value in update.items():
            existing_value = existing.get(update_key)
            if isinstance(update_value, dict) and isinstance(existing_value, dict):
                output[update_key] = merge_strategy(
                    existing_value, merge_json_elements(existing_value, update_value, merge_strategy)
                )
            else:
                merge_result = merge_strategy(existing_value, deepcopy(update_value))
                if merge_result is not None:
                    output[update_key] = merge_result

    else:
        return merge_strategy(existing, deepcopy(update))

    return output


def drop_deleted_attributes(to_be_cleaned: JsonElement, reference: JsonElement) -> JsonElement:
    """Removes all attributes from the to_be_cleaned json that are not present in the reference json."""

    # if we see a primitive type, return immediately
    if not isinstance(to_be_cleaned, (list, dict)):
        return to_be_cleaned

    # should never throw an error if the implementation is correct
    assert isinstance(to_be_cleaned, type(reference))

    # found a list, try to traverse it
    if isinstance(to_be_cleaned, list):
        reference = cast(List[JsonElement], reference)  # ensured by the assert above

        # the reference can only be smaller than to_be_cleaned if it contained env_var_strings
        # and they failed to be resolved. in that case we can use to_be_cleaned as a reference
        if len(to_be_cleaned) > len(reference):
            new_reference = to_be_cleaned
        else:
            new_reference = reference

        return [drop_deleted_attributes(tbc, ref) for tbc, ref in zip(to_be_cleaned, new_reference)]

    if isinstance(to_be_cleaned, dict):
        reference = cast(Dict[str, JsonElement], reference)

        return {k: drop_deleted_attributes(v, reference[k]) for k, v in to_be_cleaned.items() if k in reference}

    # should never happen if mypy is happy
    raise ValueError(f"Unexpected type {type(to_be_cleaned)}")


def get_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as tcp:
        tcp.bind(("", 0))
        tcp.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return tcp.getsockname()[1]  # type: ignore


def stdin_generator() -> Iterator[str]:
    if select.select([sys.stdin], [], [], 0.0)[0]:
        for line in iter(sys.stdin.readline, ""):
            yield line.rstrip("\r\n")


# makes things hashable
def freeze(elem: JsonElement) -> Any:
    # check if can be hashed first
    try:
        hash(elem)
        return elem
    except TypeError:
        pass

    if isinstance(elem, Sequence):
        return tuple([freeze(v) for v in elem])
    elif isinstance(elem, Mapping):
        return frozendict({k: freeze(v) for k, v in elem.items()})
    else:
        return elem
